"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from abc import ABC
from abc import abstractmethod
from datetime import datetime
from programy.utils.console.console import outputLog


class APIHandler(ABC):
    UNKNOWN = "Unknown"

    def __init__(self, bot_client):
        self._bot_client = bot_client

    @abstractmethod
    def process_request(self, request, method='GET'):
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def format_success_response(self, userid, question, answer, metadata=None):
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def format_error_response(self, userid, question, error, metadata=None):
        raise NotImplementedError()  # pragma: no cover


class APIHandler_V1_0(APIHandler):
    QUESTION = 'question'
    USERID = 'userid'

    def __init__(self, bot_client):
        APIHandler.__init__(self, bot_client)

    def process_request(self, client, request, method='GET'):
        question = APIHandler.UNKNOWN
        userid = APIHandler.UNKNOWN

        try:
            question = self._bot_client.get_variable(request, APIHandler_V1_0.QUESTION, method)
            userid = self._bot_client.get_variable(request, APIHandler_V1_0.USERID, method)
            client_context = client.create_client_context(userid)
            answer = self._bot_client.ask_question(userid, question)
            rendered = client.renderer.render(client_context, answer)
            return self.format_success_response(userid, question, rendered), 200

        except Exception as excep:
            return self.format_error_response(userid, question, str(excep)), 500

    def format_success_response(self, userid, question, answer, metadata=None):
        return {'response': {"question": question,
                             "answer": answer,
                             "userid": userid}}

    def format_error_response(self, userid, question, error, metadata=None):
        client_context = self._bot_client.create_client_context(userid)
        return {"response": {"question": question,
                             "answer": client_context.bot.default_response,
                             "userid": userid,
                             "error": error}}


class APIHandler_V2_0(APIHandler):
    QUERY = "query"
    USERID = 'userId'
    LANG = "lang"
    location = "location"

    def __init__(self, bot_client):
        APIHandler.__init__(self, bot_client)

    def process_request(self, client, request, method='GET'):
        query = APIHandler_V2_0.UNKNOWN
        userid = APIHandler_V2_0.UNKNOWN
        # lang =  APIHandler_V2_0.UNKNOWN
        # location =  APIHandler_V2_0.UNKNOWN

        try:
            outputLog(self, request)
            query = self._bot_client.get_variable(request, APIHandler_V2_0.QUERY, method)
            userid = self._bot_client.get_variable(request, APIHandler_V2_0.USERID, method)
            # lang = self._bot_client.get_variable(request, APIHandler_V2_0.LANG, method)
            # location = self._bot_client.get_variable(request, APIHandler_V2_0.LOCATION, method)

            metadata = {}
            answer = self._bot_client.ask_question(userid, query, metadata)
            client_context = client.create_client_context(userid)
            rendered = client.renderer.render(client_context, answer)
            return self.format_success_response(userid, query, rendered, metadata), 200

        except Exception as excep:
            return self.format_error_response(userid, query, str(excep), None), 500

    def _get_timestamp(self):
        return datetime.timestamp(datetime.now())

    def format_success_response(self, userid, question, answer, metadata=None):
        payload = {'response': {"query": question,
                                "userId": userid,
                                "timestamp": self._get_timestamp(),
                                "text": answer,
                                },
                   'status': {
                       'code': 200,
                       'message': 'success'
                   }}

        if metadata is not None:
            payload["meta"] = {}
            if 'botName' in metadata:
                payload["meta"]['botName'] = metadata['botName']
            if 'version' in metadata:
                payload["meta"]['version'] = metadata['version']
            if 'copyright' in metadata:
                payload["meta"]['copyright'] = metadata['copyright']
            if 'authors' in metadata:
                payload["meta"]['authors'] = metadata['authors']

        return payload

    def format_error_response(self, userid, question, error, metadata=None):
        payload = {'response': {"query": question,
                                "userId": userid,
                                "timestamp": self._get_timestamp(),
                                },
                   'status': {
                       'code': 500,
                       'message': error
                   }}

        if metadata is not None:
            payload["meta"] = {
                "botName": metadata['botName'],
                "version": metadata['version'],
                "copyright": metadata['copyright'],
                "authors": metadata['authors']
            }

        return payload



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import urllib.parse
import os
from programy.utils.logging.ylogger import YLogger
from programy.services.base import ServiceQuery
from programy.services.rest.base import RESTService
from programy.services.rest.base import RESTServiceException


class GNewsTopics:

    TOPICS = [
        "world",
        "nation",
        "business",
        "technology",
        "entertainment",
        "sports",
        "science",
        "health",
    ]


class GNewsSearchQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return GNewsSearchQuery(service)

    def parse_matched(self, matched):
        self._query = ServiceQuery._get_matched_var(matched, 0, "query")

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._query = None

    def execute(self):
        return self._service.search(self._query)

    def aiml_response(self, response):
        result = "SEARCH <ul>" + "".join("<li>"+x['title']+"</li>" for x in response['response']['payload']['articles']) + "</ul>"
        YLogger.debug(self, result)
        return result


class GNewsTopNewsQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return GNewsTopNewsQuery(service)

    def parse_matched(self, matched):
        self._lang = ServiceQuery._get_matched_var(matched, 0, "lang", optional=True)
        self._country = ServiceQuery._get_matched_var(matched, 1, "country", optional=True)

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._lang = None
        self._country = None

    def execute(self):
        return self._service.top_news(lang=self._lang, country=self._country)

    def aiml_response(self, response):
        result = "TOPNEWS <ul>" + "".join("<li>"+x['title']+"</li>" for x in response['response']['payload']['articles']) + "</ul>"
        YLogger.debug(self, result)
        return result


class GNewsTopicsListQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return GNewsTopicsListQuery(service)

    def parse_matched(self, matched):
        pass

    def __init__(self, service):
        ServiceQuery.__init__(self, service)

    def execute(self):
        return {'response': {'status': 'success', 'payload': {'topics': GNewsTopics.TOPICS}}}

    def aiml_response(self, response):
        result = "TOPICS <ul>" + "".join("<li>"+x+"</li>" for x in response['response']['payload']['topics']) + "</ul>"
        YLogger.debug(self, result)
        return result


class GNewsTopicQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return GNewsTopicQuery(service)

    def parse_matched(self, matched):
        self._topic = ServiceQuery._get_matched_var(matched, 0, "topic")

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._topic = None

    def execute(self):
        return self._service.topics(self._topic)

    def aiml_response(self, response):
        payload = response['response']['payload']
        articles = payload['articles']
        result = "<ul>"
        result += "\n".join(["<li>{0}</li>".format(article['title']) for article in articles])
        result += "</ul>"
        result = "TOPIC {0}".format(result)
        YLogger.debug(self, result)
        return result


class GNewsServiceException(RESTServiceException):

    def __init__(self, msg):
        RESTServiceException.__init__(self, msg)


class GNewsService(RESTService):
    """
    https://gnews.io/docs/v3#introduction
    """
    PATTERNS = [
        [r"SEARCH\s(.+)", GNewsSearchQuery],
        [r"TOPNEWS\s(.+)\s(.+)", GNewsTopNewsQuery],
        [r"TOPNEWS", GNewsTopNewsQuery],
        [r"TOPICS", GNewsTopicsListQuery],
        [r"TOPIC\s(\w+)", GNewsTopicQuery]
    ]

    SEARCH_BASE_URL=" https://gnews.io/api/v3/search"
    TOPNEWS_BASE_URL=" https://gnews.io/api/v3/top-news"
    TOPICS_BASE_URL="https://gnews.io/api/v3/topics"

    def __init__(self, configuration):
        RESTService.__init__(self, configuration)
        self._token = None

    def patterns(self) -> list:
        return GNewsService.PATTERNS

    def initialise(self, client):
        self._token = client.license_keys.get_key("GNEWS_TOKEN")
        if self._token is None:
            YLogger.error(self, "GNEWS_TOKEN missing from license.keys, service will not function correctly!")

    def get_default_aiml_file(self):
        return os.path.dirname(__file__) + os.sep + "gnews.aiml"

    @staticmethod
    def get_default_conf_file():
        return os.path.dirname(__file__) + os.sep + "gnews.conf"

    def _build_search_url(self, query, lang=None, country=None, max=None, image=None, mindate=None, maxdate=None, specific=None):
        url = GNewsService.SEARCH_BASE_URL
        url += "?token={0}".format(self._token)

        if query is not None:
            url += "&q={0}".format(urllib.parse.quote_plus(query))

        if lang is not None:
            url += "&lang={0}".format(lang)

        if country is not None:
            url += "&country={0}".format(country)

        if max is not None:
            url += "&max={0}".format(max)

        if image is not None:
            url += "&image={0}".format(image)

        if mindate is not None:
            url += "&mindate={0}".format(mindate)

        if maxdate is not None:
            url += "&maxdate={0}".format(maxdate)

        if specific is not None:
            url += "&in={0}".format(specific)

        return url

    def _build_top_news_url(self, lang=None, country=None, max=None, image=None):
        url = GNewsService.TOPNEWS_BASE_URL
        url += "?token={0}".format(self._token)

        if lang is not None:
            url += "&lang={0}".format(lang)

        if country is not None:
            url += "&country={0}".format(country)

        if max is not None:
            url += "&max={0}".format(max)

        if image is not None:
            url += "&image={0}".format(image)

        return url

    def _build_topics_url(self, topic, lang=None, country=None, max=None, image=None):
        url = GNewsService.TOPICS_BASE_URL

        url += "/{0}".format(topic)

        url += "?token={0}".format(self._token)

        if lang is not None:
            url += "&lang={0}".format(lang)

        if country is not None:
            url += "&country={0}".format(country)

        if max is not None:
            url += "&max={0}".format(max)

        if image is not None:
            url += "&image={0}".format(image)

        return url

    def search(self, query, lang=None, country=None, max=None, image=None, mindate=None, maxdate=None, specific=None):
        url = self._build_search_url(query, lang, country, max, image, mindate, maxdate, specific)
        response = self.query('search', url)
        return response

    def top_news(self, lang=None, country=None, max=None, image=None):
        if lang is not None:
            lang = lang.lower()
        if country is not None:
            country = country.lower()
        url = self._build_top_news_url(lang, country, max, image)
        response = self.query('top_news', url)
        return response

    def _validate_topic(self, topic):
        if topic not in GNewsTopics.TOPICS:
            raise GNewsServiceException("Invalid topic [%s]" % topic)

    def topics(self, topic, lang=None, country=None, max=None, image=None):
        topic = topic.lower()
        self._validate_topic(topic)
        url = self._build_topics_url(topic, lang, country, max, image)
        response = self.query('topics', url)
        return response

    def _response_to_json(self, api, response):
        return response.json()




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import random
import time


class ServiceCoordinatorConfiguration:

    def __init__(self):
        self._category = None
        self._coordinator = None
        self._services = []

    def from_yaml(self, yaml_data):
        self._category = yaml_data.get('category')
        self._coordinator = yaml_data.get('coordinator')
        self._services = yaml_data.get('services')


class ServiceCoordinator:

    def __init__(self):
        self._services = []

    def add_service(self, service):
        self._services.append(service)

    def execute_query(self, question, srai=False, skip_failures=True):
        raise NotImplementedError()


class SelectableServiceCoordinator(ServiceCoordinator):

    def __init__(self):
        ServiceCoordinator.__init__(self)

    def _get_servicen(self):
        raise NotImplementedError()

    def execute_query(self, question, srai=False, skip_failures=True, try_next_on_failure=False):
        if self._services:
            running = True
            tried = []
            while running is True:
                service = self._get_servicen()
                if service.name not in tried:
                    result = service.execite_query(question, srai)
                    if result:
                        if result['status'] == 'success':
                            return result
                        elif skip_failures is False:
                            return result

                    tried.append(service.name)

                if len(tried) == len(self._services):
                    running = False

        return None


class RandomResultServiceCoordinator(SelectableServiceCoordinator):

    def __init__(self):
        SelectableServiceCoordinator.__init__(self)

    def _get_servicen(self):
        random.seed(time.time())
        servicen = random.randint(0, len(self._services)-1)
        return self._services[servicen]


class RoundRobinResultServiceCoordinator(SelectableServiceCoordinator):

    def __init__(self):
        SelectableServiceCoordinator.__init__(self)
        self._servicen = 0

    def _get_servicen(self):
        if self._servicen >= len(self._services):
            self._servicen = 0

        service = self._services[self._servicen]
        self._servicen += 1
        return service


class AllResultsServiceCoordinator(ServiceCoordinator):

    def __init__(self):
        ServiceCoordinator.__init__(self)

    def execute_query(self, question, srai=False, skip_failures=True):
        if self._services:
            results = []
            for service in self._services:
                result = service.execite_query(question, srai)
                if result:
                    if result['status'] == 'success':
                        results.append(result)
                    elif skip_failures is False:
                        results.append(result)

            return results

        return None



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


class ServiceConfiguration:

    def __init__(self, service_type):
        self._service_type = service_type
        self._name = None
        self._category = None
        self._storage = None
        self._service_class = None
        self._default_response = None
        self._default_srai = None
        self._success_prefix = None
        self._default_aiml = None
        self._load_default_aiml = True
        self._url = None

    @staticmethod
    def from_data(service_type, name, category, storage=None, service_class=None, default_response=None, default_srai=None,
                  success_prefix=None, default_aiml=None, load_default_aiml=True, url=None):

        if service_type == 'rest':
            config = ServiceRESTConfiguration()

        elif service_type == 'wsdl':
            config = ServiceWSDLConfiguration()

        elif service_type == 'library':
            config = ServiceLibraryConfiguration()

        else:
            config = ServiceConfiguration(service_type=service_type)

        config._name = name
        config._category = category

        config._storage = storage

        config._service_class = service_class

        config._default_response = default_response
        config._default_srai = default_srai
        config._success_prefix = success_prefix

        config._default_aiml = default_aiml
        config._load_default_aiml = load_default_aiml

        config._url = url

        return config

    @property
    def service_type(self):
        return self._service_type

    @property
    def name(self):
        return self._name

    @property
    def category(self):
        return self._category

    @property
    def service_class(self):
        return self._service_class

    @property
    def storage(self):
        return self._storage

    @property
    def default_aiml(self):
        return self._default_aiml

    @property
    def load_default_aiml(self):
        return self._load_default_aiml

    @property
    def default_response(self):
        return self._default_response

    @property
    def default_srai(self):
        return self._default_srai

    @property
    def success_prefix(self):
        return self._success_prefix

    @property
    def url(self):
        return self._url

    @staticmethod
    def new_from_yaml(yaml_data, filename):

        if 'service' not in yaml_data:
            raise ValueError("'service' missing from service yaml")

        service_data = yaml_data['service']

        if 'type' in service_data:
            service_type = service_data.get('type')
            if service_type == 'rest':
                config = ServiceRESTConfiguration()

            elif service_type == 'wsdl':
                config = ServiceWSDLConfiguration()

            elif service_type == 'library':
                config = ServiceWSDLConfiguration()

            else:
                raise ValueError("Unknown service type [%s]"%service_type)

        else:
            config = ServiceConfiguration(service_type='generic')

        config.from_yaml(service_data, filename)
        return config

    def from_yaml(self, service_data, filename):

        self._name = service_data.get('name', None)
        self._category = service_data.get('category', None)

        self._storage = filename

        self._service_class = service_data.get('service_class', None)

        self._default_response = service_data.get('default_response', None)
        self._default_srai = service_data.get('default_srai', None)
        self._success_prefix = service_data.get('success_prefix', None)

        self._default_aiml = service_data.get('default_aiml', None)
        self._load_default_aiml = service_data.get('load_default_aiml', True)

        self._url = service_data.get('url', None)

    @staticmethod
    def from_sql(dao):

        if dao.type == 'rest':
            config = ServiceRESTConfiguration()

            config._retries = dao.rest_retries
            if config._retries is None:
                config._retries = ServiceRESTConfiguration.DEFAULT_RETRIES

            config._timeout = dao.rest_timeout
            if config._timeout is None:
                config._timeout = ServiceRESTConfiguration.DEFAULT_TIMEOUT

        elif dao.type == 'library':
            config = ServiceLibraryConfiguration()

        else:
            config = ServiceConfiguration(service_type=dao.type)

        config._name = dao.name
        config._category = dao.category

        config._storage = "sql"

        config._service_class = dao.service_class

        config._default_response = dao.default_response
        config._default_srai = dao.default_srai
        config._success_prefix = dao.success_prefix

        config._default_aiml = dao.default_aiml
        config._load_default_aiml = dao.load_default_aiml

        config._url = dao.url

        return config

    @staticmethod
    def from_mongo(dao):

        if dao.get('type') == 'rest':
            config = ServiceRESTConfiguration()
            rest_data = dao.get("rest", None)

            if rest_data is not None:
                config._retries = rest_data.get('retries', None)
                config._timeout = rest_data.get('timeout', None)

            if config._retries is None:
                config._retries = ServiceRESTConfiguration.DEFAULT_RETRIES

            if config._timeout is None:
                config._timeout = ServiceRESTConfiguration.DEFAULT_TIMEOUT

        elif dao.get('type')== 'library':
            config = ServiceLibraryConfiguration()

        else:
            config = ServiceConfiguration(service_type=dao.get('type'))

        config._name = dao.get('name', None)
        config._category = dao.get('category', None)

        config._storage = "mongo"

        config._service_class = dao.get('service_class', None)

        config._default_response = dao.get('default_response', None)
        config._default_srai = dao.get('default_srai', None)
        config._success_prefix = dao.get('success_prefix', None)

        config._default_aiml = dao.get('default_aiml', None)
        config._load_default_aiml = dao.get('load_default_aiml', True)

        config._url = dao.get('url', None)

        return config


class ServiceLibraryConfiguration(ServiceConfiguration):

    def __init__(self):
        ServiceConfiguration.__init__(self, service_type='library')


class ServiceRESTConfiguration(ServiceConfiguration):

    DEFAULT_RETRIES = [100, 500, 1000, 2000, 5000, 10000]
    DEFAULT_TIMEOUT = 3000

    def __init__(self):
        ServiceConfiguration.__init__(self, service_type='rest')
        self._retries = ServiceRESTConfiguration.DEFAULT_RETRIES
        self._timeout = ServiceRESTConfiguration.DEFAULT_TIMEOUT

    @property
    def retries(self):
        return self._retries

    @property
    def timeout(self):
        return self._timeout

    def from_yaml(self, service_data, filename):

        super(ServiceRESTConfiguration, self).from_yaml(service_data, filename)

        rest_data = service_data.get("rest", None)

        if rest_data is not None:

            self._retries = rest_data.get('retries', None)
            if self._retries is None:
                self._retries = ServiceRESTConfiguration.DEFAULT_RETRIES

            self._timeout = rest_data.get('timeout', None)
            if self._timeout is None:
                self._timeout = ServiceRESTConfiguration.DEFAULT_TIMEOUT


class ServiceWSDLConfiguration(ServiceConfiguration):

    def __init__(self):
        ServiceConfiguration.__init__(self, service_type='wsdl')
        self._wsdl_file = None
        self._station_codes_file = None

    @property
    def wsdl_file(self):
        return self._wsdl_file

    @property
    def station_codes_file(self):
        return self._station_codes_file

    def from_yaml(self, service_data, filename):

        super(ServiceWSDLConfiguration, self).from_yaml(service_data, filename)

        self._station_codes_file = service_data.get("station_codes_file", None)

        wsdl_data = service_data.get("wsdl", None)
        if wsdl_data is not None:
            self._wsdl_file = wsdl_data.get('wsdl_file', None)




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger

from programy.clients.client import BotClient


class EventBotClient(BotClient):

    def __init__(self, botid, argument_parser=None):
        self._running = False
        BotClient.__init__(self, botid, argument_parser=argument_parser)

    def prior_to_run_loop(self):
        pass    # pragma: no cover

    def run_loop(self):
        self._running = True
        while self._running is True:
            self._running = self.wait_and_answer()

    def wait_and_answer(self):
        raise NotImplementedError("You must override this and implement the "
                                  "logic wait for a question and send an answer back")  # pragma: no cover

    def post_run_loop(self):
        pass    # pragma: no cover

    def run(self, app=None):
        del app

        if self.arguments.noloop is False:
            YLogger.info(self, "Entering conversation loop...")

            self.startup()

            self.prior_to_run_loop()

            self.run_loop()

            self.post_run_loop()

            self.shutdown()

        else:
            YLogger.debug(self, "noloop set to True, exiting...")



import argparse
import os
import sys
import zipfile
import shutil
import wget

ybot = "https://github.com/keiffster/y-bot/archive/master.zip"
servusai = "https://github.com/keiffster/servusai-y/archive/master.zip"

programy_root = "/opt/program-y"
servusai_root = "./Servusai"
ybot_root = "./Y-bot"

systemd_root = "/etc/systemd"  # /system

webserver = "nginx"
webserver_root = "/etc/nginx"      # /site-enabled

clients = [
    {"name": "Alexa", "src": "./alexa"},
    {"name": "Console", "src": "./console"},
    {"name": "Discord", "src": "./discord"},
    {"name": "Facebook", "src": "./facebook"},
    {"name": "Google", "src": "./google"},
    {"name": "Kik", "src": "./kik"},
    {"name": "Line", "src": "./line"},
    {"name": "Microsoft", "src": "./microsoft"},
    {"name": "REST", "src": "./rest"},
    {"name": "Slack", "src": "./slack"},
    {"name": "TCP Socket", "src": "./socket"},
    {"name": "Telegram", "src": "./telegram"},
    {"name": "Twillio", "src": "./twillio"},
    {"name": "Twitter", "src": "./twitter"},
    {"name": "Viber", "src": "./viber"},
    {"name": "Web", "src": "./web", "remove_default": True},
    {"name": "XMPP", "src": "./xmpp"}
]


class Installer:

    def __init__(self):
        self._args = None
        self._name = "y-bot"

        self._programy_root = programy_root

        self._ybot = False
        self._ybot_root = ybot_root

        self._servusai = False
        self._servusai_root = servusai_root

        self._systemd_root = systemd_root

        self._webserver = webserver
        self._webserver_root = webserver_root

        self._clients = False
        self._client = None

        self._debug = False

    def parse_arguments(self):
        parser = argparse.ArgumentParser(description='Program-u multi client installer')

        parser.add_argument('--name', dest='name', help='Name of your multi client bot')

        parser.add_argument('--programyroot', dest='programyroot', help='Programy install folder')

        parser.add_argument('--ybot', dest='ybot', action='store_true', help='Install y-bot core')
        parser.add_argument('--ybotroot', dest='ybotroot', help='Y-bot install folder')

        parser.add_argument('--servusai', action='store_true', help='Install servusai clients')
        parser.add_argument('--servusairoot', dest='servusairoot', help='Servusai install folder')

        parser.add_argument('--systemdroot', dest='systemdroot', help='Systemd root directory')

        parser.add_argument('--webserver', dest='webserver', help='Which web server to install for [nginx]')
        parser.add_argument('--webserverroot', dest='webserverroot', help='Web server root folder')

        parser.add_argument('--list', dest='listclients', action='store_true', help='List all clients')
        parser.add_argument('--clients', dest='clients', action='store_true', help='Loop through install of clients')
        parser.add_argument('--client', dest='client', help='Install specific client')

        parser.add_argument('--debug', dest='debug', action='store_true', help='Do not actually execute commands')

        self._args = parser.parse_args()

        self._check_args()

        self._arg_defaults()

    def _check_args(self):
        if self._args.clients is True and self._args.client is not None:
            raise Exception("Both --clients and --client are not allowed!")

        if self._args.clients is True or self._args.client is not None:
            if self._args.webserver != "nginx":
                raise Exception("Only Nginx suppported at this time!")

    def _arg_defaults(self):
        if self._args.programyroot is not None:
            self._programy_root = self._args.programyroot

        self._servusai = self._args.servusai

        if self._args.servusairoot is not None:
            self._servusai_root = self._args.servusairoot

        self._ybot = self._args.ybot

        if self._args.ybotroot is not None:
            self._ybot_root = self._args.ybotroot

        if self._args.systemdroot is not None:
            self._systemd_root = self._args.systemdroot

        if self._args.webserver is not None:
            self._webserver = self._args.webserver

        if self._args.webserverroot is not None:
            self._webserver_root = self._args.webserverroot

        self._clients = self._args.clients

        if self._args.client is not None:
            self._client = self._args.client

        self._debug = self._args.debug

        self._print_settings()

    def _print_settings(self):
        print("Settings:")
        print("\tName:", self._name)

        print("\tProgramy Root:", self._programy_root)

        print("\tY-Bot?", self._ybot)
        print("\tY-Bot Root:", self._ybot_root)

        print("\tServusai?", self._servusai)
        print("\tServusai Root:", self._servusai_root)

        print("\tSystemd Root:", self._systemd_root)

        print("\tWeb Server:", self._webserver)
        print("\tWeb Server Root:", self._webserver_root)

        print("\tClients?", self._clients)
        if self._client is not None:
            print("\tClient:", self._client)

    def _wget_download(self, url):
        return wget.download(url, bar=wget.bar_thermometer)     # pragma: no cover

    def _download_bot(self, name, url):
        print("Downloading [%s] from [%s]" % (name, url))
        filename = self._wget_download(url)

        self._extract_bot(filename)                                                      # pragma: no cover

        master_dir = self._zip_dir_name_from_filename(filename)                          # pragma: no cover

        os.mkdir(name)

        self._recursive_copy(master_dir, name)                                           # pragma: no cover

        if os.name == 'posix':                                                           # pragma: no cover
            self._make_all_executable(name + os.sep + "scripts" + os.sep + "xnix")       # pragma: no cover

        shutil.rmtree(master_dir)                                                        # pragma: no cover

    def _extract_bot(self, filename, path=None, remove_after=True):
        zip_ref = zipfile.ZipFile(filename, 'r')
        zip_ref.extractall(path=path)
        zip_ref.close()
        if remove_after is True:
            os.remove(filename)

    def _zip_dir_name_from_filename(self, filename):
        return filename.split(".")[0]

    def _recursive_copy(self, src, dest):
        """
        Copy each file from src dir to dest dir, including sub-directories.
        """
        for item in os.listdir(src):
            file_path = os.path.join(src, item)

            # if item is a file, copy it
            if os.path.isfile(file_path):
                shutil.copy(file_path, dest)

            # else if item is a folder, recurse
            else: # os.path.isdir(file_path):
                new_dest = os.path.join(dest, item)
                os.mkdir(new_dest)
                self._recursive_copy(file_path, new_dest)

    def _make_all_executable(self, path):
        for item in os.listdir(path):
            file_path = os.path.join(path, item)
            self._make_executable(file_path)

    def _make_executable(self, path):
        mode = os.stat(path).st_mode
        mode |= (mode & 0o444) >> 2  # copy R bits to X
        os.chmod(path, mode)

    def install(self):
        if self._args.ybot is True:
            self._install_ybot()

        if self._args.servusai is True:
            self._install_servusai()

        if self._args.listclients is True:
            self._list_clients()

        if self._args.clients is True:
            self._install_clients()

        if self._args.client is not None:
            self._install_client(self._args.client)

    def _list_clients(self):
        print("\nAvailable Clients:")
        for client in clients:
            print("\t", client['name'])
        print()

    def _install_ybot(self):
        print("\nInstall Y-Bot Core")
        self._download_bot("Y-Bot", ybot)

    def _install_servusai(self):
        print("\nInstall Servusai")
        self._download_bot("Servusai", servusai)

    def _install_clients(self):
        print("\nInstalling Clients")
        for client in clients:
            self._install_client(client['name'])

    def _install_client(self, name):
        for client in clients:
            if name.upper() == client['name'].upper():
                print("\nInstalling Client", name)
                self._do_client_install(client)
                return

        raise Exception("Unknown client", name)

    def _do_client_install(self, client):

        self._install_systemd_service(client)

        self._install_web_service(client)

    def _install_systemd_service(self, client):
        print("%s systemd service install" % client['name'])

        service_file = "servusai-%s.service" % client["name"].lower()

        service_file_src = self._servusai_root + "/Servusai/%s/config/xnix/%s" % (client["name"].lower(), service_file)
        service_file_dest = self._systemd_root + "/system/%s" % service_file

        self._execute("cp %s %s" %(service_file_src, service_file_dest))

        self._execute("systemctl daemon-reload")

        self._execute("systemctl enable %s" % service_file)
        self._execute("systemctl start %s" % service_file)
        self._execute("systemctl status %s" % service_file)
        print()

    def _install_web_service(self, client):
        print("%s web service install" % client['name'])

        if client.get("remove_default", False) is True:
            self._execute("rm /etc/nginx/sites-enabled/default")

        web_site_src = self._servusai_root + "/Servusai/%s/config/xnix/%s.site" % (client["name"].lower(), client["name"].lower())
        web_site_dest = self._webserver_root + "/sites-enabled"

        self._execute("cp %s %s" % (web_site_src, web_site_dest))
        self._execute("service nginx restart")
        print()

    def _execute(self, command):
        try:
            if self._debug is False:
                stream = os.popen(command)
                output = stream.read()
                stream.close()
                print(command, "->", output)

            else:
                print(command)

            return True

        except Exception as error:
            print(error)

        return False


if __name__ == '__main__':

    installer = Installer()

    installer.parse_arguments()

    installer.install()




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import yaml
from programy.utils.logging.ylogger import YLogger
from programy.utils.classes.loader import ClassLoader
from programy.storage.entities.store import Store
from programy.storage.stores.nosql.mongo.store.mongostore import MongoStore
from programy.storage.entities.services import ServicesStore
from programy.storage.stores.nosql.mongo.dao.service import Service
from programy.services.config import ServiceConfiguration
from programy.utils.console.console import outputLog


class MongoServiceStore(MongoStore, ServicesStore):
    ServiceS = 'services'

    def __init__(self, storage_engine):
        MongoStore.__init__(self, storage_engine)
        ServicesStore.__init__(self)

    def collection_name(self):
        return MongoServiceStore.ServiceS

    def _get_entity(self, service_data):
        name = service_data.get('name')
        category = service_data.get('category')
        service_class = service_data.get('service_class')

        default_response = service_data.get('default_response')
        default_srai = service_data.get('default_srai')
        default_aiml = service_data.get('default_aiml')
        load_default_aiml = service_data.get('load_default_aiml', True)

        type = 'generic'
        rest_timeout = None
        rest_retries = None
        rest_api = None
        rest_apikey = None
        if 'rest' in service_data:
            type = 'rest'
            rest_data = service_data.get('rest', None)
            if rest_data is not None:
                rest_timeout = rest_data.get('timeout')
                rest_retries = rest_data.get('retries')
                rest_api = rest_data.get('api')
                rest_apikey = rest_data.get('apikey')

        return Service(type=type, name=name, category=category, service_class=service_class,
                 default_response=default_response, default_srai=default_srai, default_aiml=default_aiml,
                 rest_timeout=rest_timeout, rest_retries=rest_retries)

    def load(self, collector, name=None):
        YLogger.info(self, "Loading %s services from Mongo", self.collection_name())
        services = self.get_all_services()
        for service in services:
            try:
                configuration = ServiceConfiguration.from_mongo(service)
                collector.add_service(service['name'], ClassLoader.instantiate_class(service['service_class'])(configuration))

            except Exception as excep:
                YLogger.exception(self, "Failed pre-instantiating Service [%s]", excep, service['service_class'])

    def get_all_services(self):
        collection = self.collection()
        return collection.find()

    def upload_from_file(self, filename, fileformat=Store.TEXT_FORMAT, commit=True, verbose=False):

        YLogger.info(self, "Uploading %s to Mongo from file [%s]", self.collection_name(), filename)
        try:
            if self._load_services_from_file(filename, verbose) is True:
                return 1, 1

        except Exception as excep:
            YLogger.exception(self, "Error loading file [%s]", excep, filename)

        return 0, 0

    def _load_services_from_file(self, filename, verbose):

        result = False
        with open(filename, "r+") as file:
            yaml_data = yaml.load(file, Loader=yaml.FullLoader)

            service_data = yaml_data['service']
            service = self._get_entity(service_data)
            result = self.add_document(service)
            if result is True:
                if verbose is True:
                    outputLog(self, "[%s] = [%s]" % (service_data['name'], service_data['service_class']))

        return result



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import re
from abc import ABC
from abc import abstractmethod
from programy.dialog.sentence import Sentence
from programy.utils.logging.ylogger import YLogger


class ServiceException(Exception):

    def __init__(self, msg: str):
        Exception.__init__(self, msg)


class ServiceQuery(ABC):

    def __init__(self, service):
        self._service = service

    def parse_matched(self, matched):
        pass

    @abstractmethod
    def execute(self):
        raise NotImplementedError

    def aiml_response(self, response):
        YLogger.debug(self, response)
        return response

    @staticmethod
    def _get_matched_var(matched, index, name, optional=False):
        try:
            value = matched[index]

            if value is 'NONE':
                return None
            return value.strip()

        except:
            if optional is False:
                raise ValueError("query variable [%s] missing" % name)

        return None


class Service:

    DEFAULT_RESPONSE = "Service failed to return valid response"

    def __init__(self, configuration):
        self._configuration = configuration

    @property
    def configuration(self):
        return self._configuration

    @property
    def category(self) -> str:
        return self.configuration.category

    @property
    def name(self) -> str:
        return self.configuration.name

    def patterns(self) -> list:
        return []

    def initialise(self, client):
        pass

    def _match(self, question: str):
        """For each pattern for the grammar, try to find a match"""
        for pattern in self.patterns():
            matched = re.match(pattern[0], question, re.IGNORECASE)
            if matched is not None:
                query = pattern[1].create(self)
                if query is not None:
                    query.parse_matched(matched.groups())
                    YLogger.debug(self, "Service matched question [{0}] to pattern [{0}]".format(question, pattern[0]))
                    return query

        return None

    def execute_query(self, question, aiml=False):
        query = self._match(question)
        if query is not None:
            response = query.execute()
            if response is not None:
                if response['response']['status'] == 'success':
                    if aiml is True:
                        if self.configuration.success_prefix is not None:
                            result = "{0} {1}".format(self.configuration.success_prefix, query.aiml_response(response))
                        else:
                            result = query.aiml_response(response)
                    else:
                        result = response

                    YLogger.debug(self, result)
                    return result

                YLogger.error(self, "Service failed to execute query [%s] - [%s]", question, response)

        YLogger.debug(self, "Service failed to execute query [%s], empty response!", question)
        return None

    def _get_bot_response(self, client_context, sentence):
        return client_context.bot.ask_question(client_context, sentence)        # pragma: no cover

    def _get_default_response(self, client_context):
        response = None

        # Do we have an srai that we can ask the bot to expand on
        if self.configuration.default_srai is not None:
            sentence = Sentence(client_context, self.configuration.default_srai)
            response = client_context.bot.ask_question(client_context, sentence)

        # If not, do we have a default respond in the config
        if response is None:
            response = self.configuration.default_response

        # If not use the hard coded service default response
        if response is None:
            response = Service.DEFAULT_RESPONSE

        return response

    def ask_question(self, client_context, question):
        response = self.execute_query(question, aiml=True)

        # If we still don't have a response, get a default response
        if response is None:
            response = self._get_default_response(client_context)

        return response

    def get_default_aiml_file(self):
        return None

    def load_default_aiml(self, parser):

        if self.configuration.load_default_aiml is True:
            YLogger.debug(self, "Loading default aiml for service [%s]", self.name)

            filepath = None
            # First see if there is config option for default aiml
            if self.configuration.default_aiml is not None:
                filepath = self.configuration.default_aiml

            # If not then check if there is a internal name
            elif self.get_default_aiml_file() is not None:
                filepath = self.get_default_aiml_file()

            else:
                YLogger.info(self, "No default aiml for service [%s]", self.name)

            # If a file path exists, then load it into the parser
            if filepath is not None:
                return parser.parse_from_file(filepath, userid="*")

        return False



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.utils.classes.loader import ClassLoader
from programy.config.bot.joiner import BotSentenceJoinerConfiguration


class SentenceJoiner:

    def __init__(self, joiner_config):

        assert joiner_config is not None
        assert isinstance(joiner_config, BotSentenceJoinerConfiguration)

        self._configuration = joiner_config

    def combine_answers(self, answers, srai):
        final_sentences = []
        for sentence in answers:
            if sentence:

                # Capitalise the start of each sentence
                if sentence[0].isalpha():
                    sentence = sentence[0].upper() + sentence[1:]

                # If it ends with a terminator, keep the terminator, otherwise add a full stop
                if self.ends_with_terminator(sentence):
                    final_sentences.append(sentence)
                else:
                    if srai is False:
                        final_sentences.append(sentence + self._configuration.terminator)
                    else:
                        final_sentences.append(sentence)

        return " ".join([sentence for sentence in final_sentences])

    def ends_with_terminator(self, sentence):
        for ch in self._configuration.join_chars:
            if sentence.endswith(ch):
                return True
        return False

    @staticmethod
    def initiate_sentence_joiner(joiner_config):
        if joiner_config.classname is not None:
            try:
                YLogger.info(None, "Loading sentence joiner from class [%s]", joiner_config.classname)
                joiner_class = ClassLoader.instantiate_class(joiner_config.classname)
                sentence_joiner = joiner_class(joiner_config)
                return sentence_joiner
            except Exception as excep:
                YLogger.exception(None, "Failed to initiate sentence joiner", excep)
        else:
            YLogger.warning(None, "No configuration setting for sentence joiner!")

        return None



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
import json
import urllib.parse
from datetime import datetime
from programy.utils.logging.ylogger import YLogger
from programy.services.base import ServiceQuery
from programy.services.rest.base import RESTService
from programy.services.rest.base import RESTServiceException


class GetGuidelinesAllServiceQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return GetGuidelinesAllServiceQuery(service)

    def parse_matched(self, matched):
        conditions = ServiceQuery._get_matched_var(matched, 0, "conditions")
        self._conditions = ",".join([x.lower() for x in conditions.split(" ")])

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._conditions = None

    def execute(self):
        return self._service.all(self._conditions)

    def aiml_response(self, response):
        payload = response['response']['payload']
        data = payload['data']
        data2 = [x['Recommendation'] for x in data]
        recommends = ["<li>{0}</li>\n".format(x) for x in data2]
        result = 'ALL <ol>{0}</ol>'.format("".join(recommends))
        YLogger.debug(self, result)
        return result


class GetGuidelinesVacServiceQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return GetGuidelinesVacServiceQuery(service)

    def parse_matched(self, matched):
        conditions = ServiceQuery._get_matched_var(matched, 0, "conditions")
        self._conditions = ",".join([x.lower() for x in conditions.split(" ")])

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._conditions = None

    def execute(self):
        return self._service.vac(self._conditions)

    def aiml_response(self, response):
        payload = response['response']['payload']
        result = 'CONDITIONS {0}'.format(str(payload))
        YLogger.debug(self, result)
        return result


class GetGuidelinesConditionsServiceQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return GetGuidelinesConditionsServiceQuery(service)

    def __init__(self, service):
        ServiceQuery.__init__(self, service)

    def execute(self):
        return self._service.conditions()

    def aiml_response(self, response):
        payload = response['response']['payload']

        result = "<ul><li>Key - Description</li>\n"
        for key, descr in payload.items():
            result += "<li>{0} - {1}</li>\n".format(key, descr)
        result += "</ul>"

        result = 'CONDITIONS {0}'.format(result)
        YLogger.debug(self, result)
        return result


class GetGuidelinesServiceException(RESTServiceException):

    def __init__(self, msg):
        RESTServiceException.__init__(self, msg)


class GetGuidelinesService(RESTService):
    """
    """
    PATTERNS = [
        [r"ALL\s(.+)", GetGuidelinesAllServiceQuery],
        [r"VAC", GetGuidelinesVacServiceQuery],
        [r"CONDITIONS",GetGuidelinesConditionsServiceQuery]
    ]

    # conditions: Disease-specific recs. E.g. Congestive heart failure, chf Diabetes, dm. Sign in to see all conditions (too many to list here).
    CONDITIONS = {
        "dm": "Diabetes",
        "obese": "Overweight or Obesity",
        "hiv": "HIV",
        "sti": "Sexually transmitted infection",
        "postmenopause": "Postmenopause",
        "ivda": "Intravenous drug abuse",
        "msm": "Men who have sex with men",
        "hca": "Healthcare-associated",
        "liver": "Liver disease",
        "prison": "Incarceration history",
        "dialysis": "Dialysis (any type)",
        "asplenia": "Functional or anatomic asplenia",
        "scd": "Sickle cell disease or trait",
        "htn": "Hypertension",
        "heart": "Chronic heart disease",
        "lung": "Chronic lung disease",
        "csf": "Cerebrospinal fluid leak",
        "cochlear": "Cochlear implant",
        "ckd": "Chronic kidney disease",
        "ca": "Cancer / malignancy",
        "smoker": "Smoking status",
        "alcohol": "Alcohol abuse",
        "transplant": "Transplant status",
        "hypothyroid": "Hypothyroidism",
        "chf": "Congestive Heart Failure"
    }

    ALL_BASE_URL = "https://getguidelines.com/all"
    VAC_BASE_URL = "https://getguidelines.com/vac"
    
    def __init__(self, configuration):
        RESTService.__init__(self, configuration)
        self._api_token = None

    def patterns(self) -> list:
        return GetGuidelinesService.PATTERNS

    def initialise(self, client):
        self._api_token = client.license_keys.get_key('GETGUIDELINES')
        if self._api_token is None:
            YLogger.error(self, "GETGUIDELINES missing from license.keys, service will not function correctly!")

    def get_default_aiml_file(self):
        return os.path.dirname(__file__) + os.sep + "getguidelines.aiml"

    @staticmethod
    def get_default_conf_file():
        return os.path.dirname(__file__) + os.sep + "getguidelines.conf"

    def _build_all_url(self,
                   conditions,
                   bmi=None,
                   age=None,
                   pack_years=None,
                   tst=None,
                   sbp=None,
                   dbp=None,
                   ldl=None,
                   trigs=None,
                   sex=None,
                   ethnicity=None,
                   pregnant=None,
                   cac=None,
                   broad=None,
                   ):
        url = GetGuidelinesService.ALL_BASE_URL
        url += "?api_token={0}".format(self._api_token)
        url += "&conditions={0}".format(conditions)
        return url

    def _build_vac_url(self,
                   conditions,
                   bmi=None,
                   age=None,
                   pack_years=None,
                   tst=None,
                   sbp=None,
                   dbp=None,
                   ldl=None,
                   trigs=None,
                   sex=None,
                   ethnicity=None,
                   pregnant=None,
                   cac=None,
                   broad=None,
                   ):
        url = GetGuidelinesService.VAC_BASE_URL
        url += "?api_token={0}".format(self._api_token)
        url += "&conditions={0}".format(urllib.parse.quote(conditions))
        return url

    def _check_query_data(self,
                          conditions,
                          bmi,
                          age,
                          pack_years,
                          tst,
                          sbp,
                          dbp,
                          ldl,
                          trigs,
                          sex,
                          ethnicity,
                          pregnant,
                          cac,
                          broad):

        # bmi: Body mass index. Alternatively, set kg and m, or lbs and inches to have bmi calculated and applied to corresponding guidelines.
        # tst: TB skin test, measured in millimeters of induration. E.g. tst=6
        # ldl & trigs: Cholesterol profile, LDL and triglyceride.
        # pregnant: Set =1, or specificy the number of weeks pregnant for prenatal guidelines.
        # sbp: Systolic blood pressure, I.e. "The top number. So, for 135/80, sbp=135."
        # sex: Female, male: f m
        # cac: Coronary artery calcium score.
        # dbp: Diastolic blood pressure, I.e. "The bottom number. So, for 135/80, dbp=80."
        # ethnicity: African American (aa), hispanic, white.
        # broad: Flag to expand results to broad match (default is exact).

        splits_conditions = conditions.split(",")
        for condition in splits_conditions:
            if condition not in GetGuidelinesService.CONDITIONS:
                raise GetGuidelinesServiceException(
                    "Unknown condition [{0} (lowercase and no spaces!)".format(condition))

    def all(self,
            conditions,
            bmi=None,
            age=None,
            pack_years=None,
            tst=None,
            sbp=None,
            dbp=None,
            ldl=None,
            trigs=None,
            sex=None,
            ethnicity=None,
            pregnant=None,
            cac=None,
            broad=None,
            ):
        self._check_query_data(
            conditions,
            bmi,
            age,
            pack_years,
            tst,
            sbp,
            dbp,
            ldl,
            trigs,
            sex,
            ethnicity,
            pregnant,
            cac,
            broad)
        url = self._build_all_url(
            conditions,
            bmi,
            age,
            pack_years,
            tst,
            sbp,
            dbp,
            ldl,
            trigs,
            sex,
            ethnicity,
            pregnant,
            cac,
            broad)
        response = self.query('all', url)
        return response

    def vac(self,
            conditions,
            bmi=None,
            age=None,
            pack_years=None,
            tst=None,
            sbp=None,
            dbp=None,
            ldl=None,
            trigs=None,
            sex=None,
            ethnicity=None,
            pregnant=None,
            cac=None,
            broad=None,
            ):
        self._check_query_data(
            conditions,
            bmi,
            age,
            pack_years,
            tst,
            sbp,
            dbp,
            ldl,
            trigs,
            sex,
            ethnicity,
            pregnant,
            cac,
            broad)
        url = self._build_vac_url(
            conditions,
            bmi,
            age,
            pack_years,
            tst,
            sbp,
            dbp,
            ldl,
            trigs,
            sex,
            ethnicity,
            pregnant,
            cac,
            broad)
        response = self.query('vac', url)
        return response

    def conditions(self):
        started = datetime.now()
        speed = started - datetime.now()
        return self._create_success_payload("conditions", "", "DIRECT", retries=0, started=started, speed=speed, response=GetGuidelinesService.CONDITIONS)

    def _response_to_json(self, api, response):

        if api == 'conditions':
            return response

        elif api == 'all' or api == 'vac':
            return json.loads(response.text)

        return None


"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import urllib.parse
import os
from programy.utils.logging.ylogger import YLogger
from programy.services.base import ServiceQuery
from programy.services.rest.base import RESTService
from programy.services.rest.base import RESTServiceException


class CocktailDBSearchByNameServiceQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return CocktailDBSearchByNameServiceQuery(service)

    def parse_matched(self, matched):
        self._name = ServiceQuery._get_matched_var(matched, 0, "name")

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._name = None

    def execute(self):
        return self._service.search_by_name(self._name)

    def aiml_response(self, response):
        payload = response['response']['payload']
        drinks = payload['drinks']
        drink = drinks[0]

        name = drink['strDrink']
        instructions = drink['strInstructions']
        glass = drink['strGlass']

        ingredients = self._get_ingredients(drink)
        formatted = self._format_ingredients(ingredients)

        result= "NAME {0} INGREDIENTS {1} INSTRUCTIONS {2}".format(name, formatted, instructions)
        YLogger.debug(self, result)
        return result

    @staticmethod
    def _get_ingredients(drink):
        more = True
        count = 1
        ingredients = []
        while more is True and count < 16:
            ingredient_num = 'strIngredient{0}'.format(count)
            measure_num = 'strMeasure{0}'.format(count)

            ingredient = drink.get(ingredient_num, None)
            measure = drink.get(measure_num, None)

            if ingredient is not None:
                ingredients.append((measure, ingredient))
                count += 1
            else:
                more = False

        return ingredients

    @staticmethod
    def _format_ingredients(ingredients):
        return " and ".join("{0} {1}".format(x[0], x[1]) for x in ingredients)


class CocktailDBSearchByIngredientServiceQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return CocktailDBSearchByIngredientServiceQuery(service)

    def parse_matched(self, matched):
        self._ingredient = ServiceQuery._get_matched_var(matched, 0, "ingredient")

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._ingredient = None

    def execute(self):
        return self._service.search_by_ingredient(self._ingredient)

    def aiml_response(self, response):
        payload = response['response']['payload']
        drinks = payload['ingredients']
        drink = drinks[0]

        name = drink['strIngredient']
        description = drink['strDescription']

        result = "NAME {0} DESCRIPTION {1}".format(name, description)
        YLogger.debug(self, result)
        return result


class CocktailDBServiceException(RESTServiceException):

    def __init__(self, msg):
        RESTServiceException.__init__(self, msg)


class CocktailDBService(RESTService):
    """
    """
    PATTERNS = [
        [r"SEARCH\sNAME\s(.+)", CocktailDBSearchByNameServiceQuery],
        [r"SEARCH\sINGREDIENT\s(.+)", CocktailDBSearchByIngredientServiceQuery]
    ]

    SEARCH_BY_NAME_URL="https://www.thecocktaildb.com/api/json/v1/{0}/search.php?s={1}"
    SEARCH_BY_INGREDIENT_URL="https://www.thecocktaildb.com/api/json/v1/{0}/search.php?i={1}"

    def __init__(self, configuration):
        RESTService.__init__(self, configuration)
        self._api_key = None

    def patterns(self) -> list:
        return CocktailDBService.PATTERNS

    def initialise(self, client):
        self._api_key = client.license_keys.get_key('THECOCKTAILDB_APIKEY')
        if self._api_key is None:
            YLogger.error(self, "THECOCKTAILDB_APIKEY missing from license.keys, service will not function correctly!")

    def get_default_aiml_file(self):
        return os.path.dirname(__file__) + os.sep + "cocktaildb.aiml"

    @staticmethod
    def get_default_conf_file():
        return os.path.dirname(__file__) + os.sep + "cocktaildb.conf"

    def _build_search_by_name_url(self, name):
        url = CocktailDBService.SEARCH_BY_NAME_URL.format(self._api_key, name)
        return url

    def search_by_name(self, name):
        url = self._build_search_by_name_url(name)
        response = self.query('search_by_name', url)
        return response

    def _build_search_by_ingredient_url(self, ingredient):
        url = CocktailDBService.SEARCH_BY_INGREDIENT_URL.format(self._api_key, ingredient)
        return url

    def search_by_ingredient(self, ingredient):
        url = self._build_search_by_ingredient_url(ingredient)
        response = self.query('search_by_ingredient', url)
        return response

    def _response_to_json(self, api, response):
        return response.json()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
import json
from urllib.parse import quote
from xmljson import abdera
from xml.etree.ElementTree import fromstring
from programy.utils.logging.ylogger import YLogger
from programy.services.base import ServiceQuery
from programy.services.rest.base import RESTService
from programy.services.rest.base import RESTServiceException


class GoodReadsSearchAuthorServiceQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return GoodReadsSearchAuthorServiceQuery(service)

    def parse_matched(self, matched):
        self._name = ServiceQuery._get_matched_var(matched, 0, "name")

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._name = None

    def execute(self):
        return self._service.search_for_author(self._name)

    def aiml_response(self, response):
        payload = response['response']['payload']
        goodreads = payload['GoodreadsResponse']
        children = goodreads['children'][1]
        author = children['author']
        attributes = author['attributes']
        id = attributes['id']
        result = "AUTHOR SEARCH {0}".foramt(str(id))
        YLogger.debug(self, result)
        return result


class GoodReadsListBooksServiceQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return GoodReadsListBooksServiceQuery(service)

    def parse_matched(self, matched):
        self._authorid = ServiceQuery._get_matched_var(matched, 0, "authorid")

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._authorid = None

    def execute(self):
        return self._service.list_books(self._authorid)

    def aiml_response(self, response):
        payload = response['response']['payload']
        goodReadsResponse = payload['GoodreadsResponse']
        children = goodReadsResponse['children']
        author = children[1]['author']
        books = author['children'][3]['books']
        titles = []
        for child in books['children']:
            book = child['book']
            for item in book['children']:
                if 'title_without_series' in item:
                    titles.append(item['title_without_series'])
                    break

        result = "BOOK LIST <ul>{0}</ul>".format("".join("<li>{0}</li>".format(x) for x in titles))
        YLogger.debug(self, result)
        return result


class GoodReadsServiceException(RESTServiceException):

    def __init__(self, msg):
        RESTServiceException.__init__(self, msg)


class GoodReadsService(RESTService):
    """
    """
    PATTERNS = [
        [r"SEARCH\sAUTHOR\s(.+)", GoodReadsSearchAuthorServiceQuery],
        [r"LIST\sBOOKS\s(.+)", GoodReadsListBooksServiceQuery]
    ]

    SEARCH_AUTHOR_URL="https://www.goodreads.com/api/author_url/{0}?key={1}"
    BOOKS_LIST_URL="https://www.goodreads.com/author/list/{0}?format=xml&key={1}"

    def __init__(self, configuration):
        RESTService.__init__(self, configuration)
        self._api_key = None
        self._api_secret = None

    def patterns(self) -> list:
        return GoodReadsService.PATTERNS

    def initialise(self, client):
        self._api_key = client.license_keys.get_key('GOODREADS_KEY')
        if self._api_key is None:
            YLogger.error(self, "GOODREADS_KEY missing from license.keys, service will not function correctly!")

        self._api_secret = client.license_keys.get_key('GOODREAD_SECRET')
        if self._api_secret is None:
            YLogger.error(self, "GOODREAD_SECRET missing from license.keys, service will not function correctly!")

    def get_default_aiml_file(self):
        return os.path.dirname(__file__) + os.sep + "goodreads.aiml"

    @staticmethod
    def get_default_conf_file():
        return os.path.dirname(__file__) + os.sep + "goodreads.conf"

    def _build_author_search_url(self, name):
        url = GoodReadsService.SEARCH_AUTHOR_URL.format(quote(name), self._api_key)
        return url

    def search_for_author(self, name):
        url = self._build_author_search_url(name)
        response = self.query('search_for_author', url)
        return response

    def _build_list_books_url(self, authorid):
        url = GoodReadsService.BOOKS_LIST_URL.format(authorid, self._api_key)
        return url

    def list_books(self, authorid):
        url = self._build_list_books_url(authorid)
        response = self.query('list_books', url)
        return response

    def _response_to_json(self, api, response):
        jsondata = abdera.data(fromstring(response.content.decode('UTF-8')))
        return json.loads(json.dumps(jsondata))

    @staticmethod
    def get_author_id(payload):
        response = payload.get("response", None)
        if response is not None:
            payload = response.get("payload", None)
            if payload is not None:
                goodreadsResponse = payload.get("GoodreadsResponse", None)
                if goodreadsResponse is not None:
                    children = goodreadsResponse.get('children', [])
                    for child in children:
                        if child.get('author', None) is not None:
                            author = child['author']
                            attributes = author.get("attributes", {})
                            return attributes.get('id', None)
        return None

    @staticmethod
    def get_books(payload):
        book_list = []
        response = payload.get("response", None)
        if response is not None:
            payload = response.get("payload", None)
            if payload is not None:
                goodreadsResponse = payload.get("GoodreadsResponse", None)
                if goodreadsResponse is not None:
                    children = goodreadsResponse.get('children', [])
                    for child in children:
                        if child.get('author', None) is not None:
                            author = child['author']
                            author_children = author.get('children', [])
                            for author_child in author_children:
                                books = author_child.get('books', {})
                                book_children = books.get('children', [])
                                for book_child in book_children:
                                    book = book_child.get('book', {})
                                    book_subchildren = book.get('children', [])
                                    book_id = book_title = None
                                    for book_subchild in book_subchildren:
                                        if 'id' in book_subchild:
                                            id_dict = book_subchild.get('id', {})
                                            if id_dict:
                                                id_dict_childs = id_dict.get('children', [])
                                                if id_dict_childs:
                                                    book_id = id_dict_childs[0]

                                        if 'title' in book_subchild:
                                            book_title = book_subchild.get('title', None)

                                        if book_id and book_title:
                                            book_list.append((book_id, book_title))
                                            break

        return tuple(book_list)




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import re
from programy.utils.logging.ylogger import YLogger


class Substitutions:
    PATTERN = re.compile(r"(\$[A-Z]+)")

    def __init__(self):
        self._substitutes = {}

    def empty(self):
        self._substitutes.clear()

    def add_substitute(self, name, value):
        if name in self._substitutes:
            YLogger.warning(self, "Substitute [%s], already exists", name)
        self._substitutes[name] = value

    def has_substitute(self, name):
        return bool(name in self._substitutes)

    def get_substitute(self, name):
        if name in self._substitutes:
            return self._substitutes[name]
        else:
            raise ValueError("No substitute named [%s]" % name)

    def load_substitutions(self, substitutions):
        if substitutions is not None:
            try:
                with open(substitutions, "r") as subsfile:
                    for line in subsfile:
                        namevalue = line.strip().split(":")

                        if len(namevalue) > 1:
                            self.add_substitute(namevalue[0], ":".join(namevalue[1:]))

            except Exception as excep:
                YLogger.warning(None, "No substitutions file, but one specified [%s]", excep)

    def replace(self, string: str) -> str:
        splits = Substitutions.PATTERN.split(string)
        replaced = []
        for segment in splits:
            if self.has_substitute(segment):
                replaced.append(self.get_substitute(segment))

            else:
                replaced.append(segment)

        return "".join(replaced)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

# Force python XML parser not faster C accelerators
# because we can't hook the C implementation
import sys  # pylint: disable=wrong-import-position

sys.modules['_elementtree'] = None
import xml.etree.ElementTree as ET  # pylint: disable=wrong-import-position


class LineNumberingParser(ET.XMLParser):
    # pylint: disable=protected-access
    def _end(self, *args, **kwargs):  # pylint: disable=arguments-differ
        superend = super(self.__class__, self)  # pylint: disable=attribute-error, bad-super-call           # pragma: no cover
        element = superend._end(*args, **kwargs)  # pytype: disable=attribute-error                         # pragma: no cover
        element._end_line_number = self.parser.CurrentLineNumber  # pylint: disable=protected-access        # pragma: no cover
        element._end_column_number = self.parser.CurrentColumnNumber  # pylint: disable=protected-access    # pragma: no cover
        element._end_byte_index = self.parser.CurrentByteIndex  # pylint: disable=protected-access          # pragma: no cover
        return element                                                                                      # pragma: no cover

    # pylint: disable=protected-access
    def _start(self, *args, **kwargs):  # pylint: disable=arguments-differ
        # Here we assume the default XML parser which is expat
        # and copy its element position attributes into output Elements
        superstart = super(self.__class__, self) # pylint: disable=attribute-error, bad-super-call           # pragma: no cover
        element = superstart._start(*args, **kwargs)  # pytype: disable=attribute-error                      # pragma: no cover
        element._start_line_number = self.parser.CurrentLineNumber  # pylint: disable=protected-access       # pragma: no cover
        element._start_column_number = self.parser.CurrentColumnNumber  # pylint: disable=protected-access   # pragma: no cover
        element._start_byte_index = self.parser.CurrentByteIndex  # pylint: disable=protected-access         # pragma: no cover
        return element                                                                                       # pragma: no cover



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from abc import ABC
import requests
import threading
from requests.exceptions import HTTPError
from datetime import datetime
from programy.utils.logging.ylogger import YLogger
from programy.services.base import Service
from programy.services.base import ServiceException


class RESTServiceException(ServiceException):

    def __init__(self, msg):
        ServiceException.__init__(self, msg)


class RESTService(Service, ABC):

    DEFAULT_RETRIES = [100, 500, 1000, 2000, 5000, 10000]

    def __init__(self, configuration):
        Service.__init__(self, configuration)

    def _response_to_json(self, api, response):
        raise NotImplementedError()     # pragma: no cover

    def _add_base_payload(self, data, status, api, url, call, retries, started, speed):
        data['response']['status'] = status
        data['response']['url'] = url
        data['response']['api'] = api
        data['response']['call'] = call
        data['response']['retries'] = retries
        if started is not None:
            data['response']['started'] = started.strftime("%d/%m/%Y, %H:%M:%S")
        if speed is not None:
            data['response']['speed'] = str(speed.microseconds/1000) + "ms"
        data['response']['service'] = self.name
        data['response']['category'] = self.category

    def _create_success_payload(self, api, url, call, retries, started, speed, response):
        data = {}
        data['response'] = {}
        self._add_base_payload(data, "success", api, url, call, retries, started, speed)
        data['response']['payload'] = self._response_to_json(api, response)
        return data

    def _create_statuscode_failure_payload(self, api, url, call, retries, started, speed, response):
        data = {}
        data['response'] = {}
        self._add_base_payload(data, "failure", api, url, call, retries, started, speed)
        data['response']['payload'] = {}
        data['response']['payload']['type'] = 'statusCode'
        data['response']['payload']['statusCode'] = response.status_code
        return data

    def _create_http_failure_payload(self, api, url, call, retries, started, speed, http_err):
        data = {}
        data['response'] = {}
        self._add_base_payload(data, "failure", api, url, call, retries, started, speed)
        data['response']['payload'] = {}
        data['response']['payload']['type'] = 'http'
        data['response']['payload']['httpError'] = str(http_err)
        return data

    def _create_general_failure_payload(self, api, url, call, retries, started, speed, err):
        data = {}
        data['response'] = {}
        self._add_base_payload(data, "failure", api, url, call, retries, started, speed)
        data['response']['payload'] = {}
        data['response']['payload']['type'] = 'general'
        data['response']['payload']['error'] = str(err)
        return data

    def _requests_get(self, url, headers, timeout):
        response = requests.get(url, headers=headers, timeout=timeout)
        return response

    def _do_get(self, url, headers=None):

        response = self._requests_get(url, headers=headers, timeout=self.configuration.timeout)
        if response.status_code != 408:
            return response, 0

        count = 1
        while count < len(self.configuration.retries):
            YLogger.warning(self, "Timeout, sleeping for {0}ms".format((self.configuration.retries[count])))
            threading.sleep(self.configuration.retries[count])

            response = self._requests_get(url, headers=headers, timeout=self.configuration.timeout)
            YLogger.debug(self, response)
            if response.status_code != 408:
                break

            count += 1

        return response, count

    def _get(self, api, url, headers=None):
        started = None
        speed = None
        retries = 0
        try:
            started = datetime.now()
            if self.configuration.timeout:
                response, retries = self._do_get(url, headers)
            else:
                response = requests.get(url, headers)
            speed = started - datetime.now()

            # If the response was successful, no Exception will be raised
            response.raise_for_status()

            if response.status_code == 200:
                return self._create_success_payload(api, url, "GET", retries, started, speed, response)

            else:
                return self._create_statuscode_failure_payload(api, url, "GET", retries, started, speed, response)

        except HTTPError as http_err:
            return self._create_http_failure_payload(api, url, "GET", retries, started, speed, http_err)

        except Exception as err:
            return self._create_general_failure_payload(api, url, "GET", retries, started, speed, err)

    def _requests_post(self, url, headers, params, timeout):
        return requests.post(url, headers=headers, data=params, timeout=timeout)

    def _do_post(self, url, params, headers=None):

        response = self._requests_post(url, headers=headers, params=params, timeout=self.configuration.timeout)
        if response.status_code != 408:
            return response, 0

        count = 1
        while count < len(self.configuration.retries):
            YLogger.warning(self, "Timeout, sleeping for {0}ms".format((self.configuration.retries[count])))
            threading.sleep(self.configuration.retries[count])

            response = self._requests_post(url, headers=headers, data=params, timeout=self.configuration.timeout)
            if response.status_code != 408:
                break

            count += 1

        return response. count

    def _post(self, api, url, params, headers=None):
        started = None
        speed = None
        retries = 0
        try:
            started = datetime.now()
            if self.configuration.timeout:
                response, retries = self._do_post(url, params, headers)
            else:
                response = requests.post(url, headers=headers, data=params)
            speed = started - datetime.now()

            # If the response was successful, no Exception will be raised
            response.raise_for_status()

            if response.status_code == 200:
                return self._create_success_payload(api, url, "post", retries, started, speed, response)

            else:
                return self._create_statuscode_failure_payload(api, url, "post", retries, started, speed, response)

        except HTTPError as http_err:
            return self._create_http_failure_payload(api, url, "post", retries, started, speed, http_err)

        except Exception as err:
            return self._create_general_failure_payload(api, url, "post", retries, started, speed, err)
            
    def query(self, api, url, type="GET", params=None, headers=None):

        if type == "GET":
            return self._get(api, url, headers=headers)

        elif type == "POST":
            return self._post(api, url, params, headers=headers)

        else:
            raise RESTServiceException("Invalid REST call type [%s]"%(type))




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
import metoffer
from datetime import datetime
from programy.services.library.metoffice.metoffice import MetOffice
from programy.utils.logging.ylogger import YLogger
from programy.services.base import ServiceQuery
from programy.services.library.base import PythonAPIService
from programy.services.utils.geocode import GeoCodeUtils


class MetOfficeObservationQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return MetOfficeObservationQuery(service)

    def parse_matched(self, matched):
        latsign = ServiceQuery._get_matched_var(matched, 0, "latsign")
        latdec = ServiceQuery._get_matched_var(matched, 1, "latdec")
        latfrac = ServiceQuery._get_matched_var(matched, 2, "latfrac")
        self._lat = GeoCodeUtils.aiml_string_to_float(latsign, latdec, latfrac)

        lngsign = ServiceQuery._get_matched_var(matched, 3, "lngsign")
        lngdec = ServiceQuery._get_matched_var(matched, 4, "lngdec")
        lngfrac = ServiceQuery._get_matched_var(matched, 5, "lngfrac")
        self._lng = GeoCodeUtils.aiml_string_to_float(lngsign, lngdec, lngfrac)

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._lat = None
        self._lng = None

    def execute(self):
        return self._service.observation(self._lat, self._lng)

    def aiml_response(self, response):
        observation = MetOffice.parse_observation(response['response']['payload']['observation'])
        current = observation.get_latest_observation()
        result = current.to_program_y_text()
        YLogger.debug(self, result)
        return result


class MetOfficeHoursForecastQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return MetOfficeHoursForecastQuery(service)

    def parse_matched(self, matched):
        latsign = ServiceQuery._get_matched_var(matched, 0, "latsign")
        latdec = ServiceQuery._get_matched_var(matched, 1, "latdec")
        latfrac = ServiceQuery._get_matched_var(matched, 2, "latfrac")
        self._lat = GeoCodeUtils.aiml_string_to_float(latsign, latdec, latfrac)

        lngsign = ServiceQuery._get_matched_var(matched, 3, "lngsign")
        lngdec = ServiceQuery._get_matched_var(matched, 4, "lngdec")
        lngfrac = ServiceQuery._get_matched_var(matched, 5, "lngfrac")
        self._lng = GeoCodeUtils.aiml_string_to_float(lngsign, lngdec, lngfrac)

        self._hours = int(ServiceQuery._get_matched_var(matched, 6, "hours"))

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._lat = None
        self._lng = None
        self._hours = None

    def execute(self):
        return self._service.forecast(self._lat, self._lng, metoffer.THREE_HOURLY)

    def _get_current_date(self):
        time_now = datetime.now()
        return time_now.strftime("%Y-%m-%dZ")

    def aiml_response(self, response):
        forecasts = MetOffice.parse_forecast(response['response']['payload']['forecast'], metoffer.THREE_HOURLY)
        forecast = forecasts.get_forecast_for_n_hours_ahead(self._hours, fromdate=self._get_current_date())
        YLogger.debug(self, forecast)
        return forecast


class MetOfficeDaysForecastQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return MetOfficeDaysForecastQuery(service)

    def parse_matched(self, matched):
        latsign = ServiceQuery._get_matched_var(matched, 0, "latsign")
        latdec = ServiceQuery._get_matched_var(matched, 1, "latdec")
        latfrac = ServiceQuery._get_matched_var(matched, 2, "latfrac")
        self._lat = GeoCodeUtils.aiml_string_to_float(latsign, latdec, latfrac)

        lngsign = ServiceQuery._get_matched_var(matched, 3, "lngsign")
        lngdec = ServiceQuery._get_matched_var(matched, 4, "lngdec")
        lngfrac = ServiceQuery._get_matched_var(matched, 5, "lngfrac")
        self._lng = GeoCodeUtils.aiml_string_to_float(lngsign, lngdec, lngfrac)

        self._days = int(ServiceQuery._get_matched_var(matched, 6, "days"))

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._lat = None
        self._lng = None
        self._days = None

    def execute(self):
        return self._service.forecast(self._lat, self._lng, metoffer.DAILY)

    def _get_current_date(self):
        time_now = datetime.now()
        return time_now.strftime("%Y-%m-%dZ")

    def aiml_response(self, response):
        forecasts = MetOffice.parse_forecast(response['response']['payload']['forecast'], metoffer.DAILY)
        forecast = forecasts.get_forecast_for_n_days_ahead(self._days, fromdate=self._get_current_date())
        YLogger.debug(self, forecast)
        return forecast


class MetOfficeService(PythonAPIService):

    PATTERNS = [
        [r"OBSERVATION\sLAT\sSIGN\s(.+)\sDEC\s(.+)\sFRAC\s(.+)\sLNG\sSIGN\s(.+)\sDEC\s(.+)\sFRAC\s(.+)", MetOfficeObservationQuery],
        [r"FORECAST\sLAT\sSIGN\s(.+)\sDEC\s(.+)\sFRAC\s(.+)\sLNG\sSIGN\s(.+)\sDEC\s(.+)\sFRAC\s(.+)\sHOURS\s(.+)", MetOfficeHoursForecastQuery],
        [r"FORECAST\sLAT\sSIGN\s(.+)\sDEC\s(.+)\sFRAC\s(.+)\sLNG\sSIGN\s(.+)\sDEC\s(.+)\sFRAC\s(.+)\sDAYS\s(.+)", MetOfficeDaysForecastQuery]
    ]

    def __init__(self, configuration):
        PythonAPIService.__init__(self, configuration)
        self._met_office = None

    def patterns(self) -> list:
        return MetOfficeService.PATTERNS

    def get_default_aiml_file(self):
        return os.path.dirname(__file__) + os.sep + "metoffice.aiml"

    @staticmethod
    def get_default_conf_file():
        return os.path.dirname(__file__) + os.sep + "metoffice.conf"

    def _response_to_json(self, api, response):
        return response

    def initialise(self, client):
        self._api_key = client.license_keys.get_key("METOFFICE_API_KEY")
        if self._api_key is None:
            YLogger.error(self, "METOFFICE_API_KEY missing from license.keys, service will not function correctly!")
        self._met_office = MetOffice(self._api_key)

    def observation(self, lat, long):
        started = datetime.now()
        speed = None
        try:
            data = self._met_office.current_observation(lat, long)
            speed = started - datetime.now()

            if data is not None:
                result = {"observation": data}
                return self._create_success_payload("observation", started, speed, result)

            return self._create_failure_payload("observation", started, speed)

        except Exception as error:
            return self._create_exception_failure_payload("observation", started, speed, error)

    def forecast(self, lat, long, forecast):
        started = datetime.now()
        speed = None
        try:
            if forecast == metoffer.DAILY:
                data = self._met_office.five_day_forecast(lat, long)

            elif forecast == metoffer.THREE_HOURLY:
                data = self._met_office.twentyfour_hour_forecast(lat, long)

            else:
                raise ValueError("Invalid forecast type")

            speed = started - datetime.now()

            if data is not None:
                result = {"forecast": data}
                return self._create_success_payload("forecast", started, speed, result)

            return self._create_failure_payload("forecast", started, speed)

        except Exception as error:
            return self._create_exception_failure_payload("forecast", started, speed, error)




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import Dict
import requests
from programy.utils.logging.ylogger import YLogger
from programy.triggers.config import TriggerConfiguration
from programy.context import ClientContext
from programy.triggers.manager import TriggerManager


class RestTriggerManager(TriggerManager):

    def __init__(self, config: TriggerConfiguration):
        TriggerManager.__init__(self, config)

    def post_data(self, api_url_base, headers, payload):
        return requests.post(api_url_base, headers=headers, json=payload)   # pragma: no cover

    def trigger(self, event: str, client_context: ClientContext = None, additional: Dict[str, str] = None) -> bool:

        if client_context is not None:
            assert isinstance(client_context, ClientContext)

        YLogger.debug(client_context, "Event triggered [%s]", event)

        payload = {'event': event}

        if client_context is not None:
            conversation = client_context.bot.get_conversation(client_context)
            payload['conversation'] = conversation.to_json()

        if additional is not None:
            payload['additional'] = additional

        api_url_base = self._config.value("url")
        api_method = self._config.value("method")
        api_token = self._config.value("token")

        headers = {'Content-Type': 'application/json'}
        if api_token is not None:
            headers['Authorisation'] = 'Bearer %s' % api_token

        if api_method is None or api_method.upper() == 'POST':
            try:
                response = self.post_data(api_url_base, headers=headers, payload=payload)

                if response.status_code == 200:
                    return True

                YLogger.error(None, "Failed to send trigger info via REST [%d]", response.status_code)

            except Exception as error:
                YLogger.exception(client_context, "Failed to post data", error)

        return False



"""
Copyright (c) 2016-2018 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

This is an example extension that allow syou to call an external service to retreive the bank balance
of the customer. Currently contains no authentication
"""
import os
from datetime import datetime
from fuzzywuzzy import process
from programy.utils.logging.ylogger import YLogger
from programy.services.config import ServiceWSDLConfiguration
from programy.services.base import ServiceQuery
from programy.services.wsdl.base import WSDLService
from programy.services.wsdl.base import WSDLServiceException


class NationalRailEnquiriesWSDLServiceConfiguration(ServiceWSDLConfiguration):

    def __init__(self, service_type='wsdl'):
        ServiceWSDLConfiguration.__init__(self, service_type)
        self._station_codes_file = None

    @property
    def station_codes_file(self):
        return self._station_codes_file

    def from_yaml(self, service_data, filename):

        super(NationalRailEnquiriesWSDLServiceConfiguration, self).from_yaml(service_data, filename)

        self._station_codes_file = service_data.get("station_codes_file", None)


class NationalRailEnquiriesStationNameServiceQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return NationalRailEnquiriesStationNameServiceQuery(service)

    def parse_matched(self, matched):
        self._crs = ServiceQuery._get_matched_var(matched, 0, "crs")

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._crs = None

    def execute(self):
        return self._service.get_station_name_from_code(self._crs)

    def aiml_response(self, response):
        payload = response['response']['payload']
        station_name= payload['station_name']
        YLogger.debug(self, station_name)
        return station_name


class NationalRailEnquiriesStationCodeServiceQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return NationalRailEnquiriesStationCodeServiceQuery(service)

    def parse_matched(self, matched):
        self._name = ServiceQuery._get_matched_var(matched, 0, "name")

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._name = None

    def execute(self):
        return self._service.get_station_code_from_name(self._name)

    def aiml_response(self, response):
        payload = response['response']['payload']
        station_code = payload['station_code']
        YLogger.debug(self, station_code)
        return station_code


class NationalRailEnquiriesNextTrainFromServiceQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return NationalRailEnquiriesNextTrainFromServiceQuery(service)

    def parse_matched(self, matched):
        self._station = ServiceQuery._get_matched_var(matched, 0, "station")
        self._platform = ServiceQuery._get_matched_var(matched, 1, "platform")
        self._destination = ServiceQuery._get_matched_var(matched, 2, "destination", optional=True)

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._station = None
        self._platform = None
        self._destination = None

    def execute(self):
        return self._service.next_trains_from_station(self._station, platform=self._platform, destination=self._destination)

    def aiml_response(self, response):
        payload = response['response']['payload']
        departures = payload['departures']
        platforms = departures['platforms']

        data = platforms[self._platform]
        first = data[0]
        std = first['std']
        eta = first['eta']
        operator = first['operator']
        origin = first['origin']
        destination = first['destination']

        data = "OPERATOR %s STD %s ETA %s ORIGIN %s DESTINATION %s"% (operator, std, eta, origin, destination)
        YLogger.debug(self, data)
        return data


class NationalRailEnquiriesWSDLServiceException(WSDLServiceException):

    def __init__(self, msg):
        WSDLServiceException.__init__(self, msg)


class NationalRailEnquiriesWSDLService(WSDLService):

    # https://lite.realtime.nationalrail.co.uk/OpenLDBWS/

    WSDL = 'https://lite.realtime.nationalrail.co.uk/OpenLDBWS/wsdl.aspx?ver=2017-10-01'

    PATTERNS = [
        [r"STATION\sNAME\s(.+)", NationalRailEnquiriesStationNameServiceQuery],
        [r"STATION\sCODE\s(.+)", NationalRailEnquiriesStationCodeServiceQuery],
        [r"NEXT\sTRAIN\sFROM\s(.+)\sPLATFORM\s(.+)\sTO\s(.+)", NationalRailEnquiriesNextTrainFromServiceQuery],
        [r"NEXT\sTRAIN\sFROM\s(.+)\sPLATFORM\s(.+)", NationalRailEnquiriesNextTrainFromServiceQuery]
    ]

    def __init__(self, configuration):
        WSDLService.__init__(self, configuration)
        self._stations_cache = []
        self._stations_to_codes = {}
        self._codes_to_stations = {}
        self._header = None

    def patterns(self) -> list:
        return NationalRailEnquiriesWSDLService.PATTERNS

    def initialise(self, client):
        access_token = client.license_keys.get_key('NATIONAL_RAIL_ENQUIRIES')
        if access_token is None:
            YLogger.error(self, "NATIONAL_RAIL_ENQUIRIES missing from license.keys, service will not function correctly!")

        self._header = self._create_header(access_token)

        if self.configuration.station_codes_file:
            self._load_station_codes_to_stations(self.configuration.station_codes_file)

        if self.configuration.wsdl_file is None:
            self.create_wsdl_client(NationalRailEnquiriesWSDLService.WSDL)
        else:
            self.create_wsdl_client(self.configuration.wsdl_file)

    def get_default_aiml_file(self):
        return os.path.dirname(__file__) + os.sep + "nationalrailenquiries.aiml"

    @staticmethod
    def get_default_conf_file():
        return os.path.dirname(__file__) + os.sep + "nationalrailenquiries.conf"

    def _load_station_codes_to_stations(self, filename):
        self._stations_to_codes.clear()
        self._codes_to_stations.clear()
        self._stations_cache.clear()
        try:
            if os.path.exists(filename) is False:
                filename = os.path.dirname(__file__) + os.sep + filename

            with open(filename, 'r') as code_file:
                for line in code_file:
                    if line and ',' in line:
                        parts = line.strip().split(",")
                        station = parts[0].upper()
                        code = parts[1].upper()
                        self._stations_to_codes[code] = station
                        self._codes_to_stations[station] = code
                        self._stations_cache.append(station)
            YLogger.debug(self, "Loaded %d stations"%len(self._stations_cache))
        except Exception as e:
            YLogger.debug(self, "Failed to load station codes from %s"%filename)

    def _create_header(self, access_token):
        return {"AccessToken": access_token}

    def _validate_num_rows(self, num_rows):
        if num_rows < 1 or num_rows > 150:
            raise AttributeError("Invalid value for num_rows, between 1 and 150 only!")

    def _validate_crs(self, crs):
        if len(crs) != 3:
            raise AttributeError("Invalid crs value 3 letters only!")

    def _validate_filterCrs(self, filterCrs):
        if len(filterCrs) != 3:
            raise AttributeError("Invalid filterCrs value, 3 letters only!")

    def _validate_filterType(self, filterType):
        if filterType not in ['from', 'to']:
            raise AttributeError("Invalid filterType value, 'to' or 'from' only!")

    def _validate_filterList(self, filterList):
        if len(filterList) < 1 or len(filterList) > 15:
            raise AttributeError("Invalid filterList value, between 1 and 15 items")
        for filterCrs in filterList:
            if len(filterCrs) != 3:
                raise AttributeError("Invalid filterCrs value in filterList, 3 letters only!")

    def _validate_timeOffset(self, timeOffset):
        if timeOffset < -120 or timeOffset > 120:
            raise AttributeError("Invalid timeOffset value, between -120 and 120 only!")

    def _validate_timeWindow(self, timeWindow):
        if timeWindow < -120 or timeWindow > 120:
            raise AttributeError("Invalid timeWindow value, between -120 and 120 only!")

    def _validate_serviceID(self, serviceID):
        if not serviceID:
            raise AttributeError("Invalid serviceID value, None not allowed!")

    # Returns all public departures for the supplied CRS code within a defined time window, including service details.
    # GetArrBoardWithDetails(
    #    numRows: xsd:unsignedShort, crs: ns2:CRSType, filterCrs: ns2:CRSType, filterType: ns2:FilterType, timeOffset: xsd:int, timeWindow: xsd:int, _soapheaders = {
    #    AccessToken: ns0:AccessToken}) -> GetStationBoardResult: ns4:StationBoardWithDetails
    def get_arrival_boards_with_details(self, numRows, crs, filterCrs=None, filterType=None, timeOffset=0, timeWindow=120):

        self._validate_num_rows(numRows)
        self._validate_crs(crs)
        if filterCrs is not None:
            self._validate_filterCrs(filterCrs)
        if filterType is not None:
            self._validate_filterType(filterType)
        self._validate_timeOffset(timeOffset)
        self._validate_timeWindow(timeWindow)

        started = datetime.now()
        speed = None
        try:
            data = self._client.service.GetArrBoardWithDetails(numRows=numRows,
                                                               crs=crs,
                                                               filterCrs=filterCrs,
                                                               filterType=filterType,
                                                               timeOffset=timeOffset,
                                                               timeWindow=timeWindow,
                                                               _soapheaders=self._header)

            speed = started - datetime.now()

            if data is not None:
                result = {"arrival_boards_with_details": data}
                return self._create_success_payload("arrival_boards_with_details", started, speed, result)

            return self._create_failure_payload("arrival_boards_with_details", started, speed)

        except Exception as error:
            return self._create_exception_failure_payload("arrival_boards_with_details", started, speed, error)

    # Returns all public arrivals and departures for the supplied CRS code within a defined time window, including service details.
    # GetArrDepBoardWithDetails(
    #    numRows: xsd:unsignedShort, crs: ns2:CRSType, filterCrs: ns2:CRSType, filterType: ns2:FilterType, timeOffset: xsd:int, timeWindow: xsd:int, _soapheaders = {
    #    AccessToken: ns0:AccessToken}) -> GetStationBoardResult: ns4:StationBoardWithDetails
    def get_arrival_and_departure_boards_with_details(self, numRows, crs, filterCrs=None, filterType=None, timeOffset=0, timeWindow=120):

        self._validate_num_rows(numRows)
        self._validate_crs(crs)
        if filterCrs is not None:
            self._validate_filterCrs(filterCrs)
        if filterType is not None:
            self._validate_filterType(filterType)
        self._validate_timeOffset(timeOffset)
        self._validate_timeWindow(timeWindow)

        started = datetime.now()
        speed = None
        try:
            data = self._client.service.GetArrDepBoardWithDetails(numRows=numRows,
                                                                  crs=crs,
                                                                  filterCrs=filterCrs,
                                                                  filterType=filterType,
                                                                  timeOffset=timeOffset,
                                                                  timeWindow=timeWindow,
                                                                  _soapheaders=self._header)

            speed = started - datetime.now()

            if data is not None:
                result = {"arrival_and_departure_boards_with_details": data}
                return self._create_success_payload("arrival_and_departure_boards_with_details", started, speed, result)

            return self._create_failure_payload("arrival_and_departure_boards_with_details", started, speed)

        except Exception as error:
            return self._create_exception_failure_payload("arrival_and_departure_boards_with_details", started, speed, error)

    # Returns all public arrivals and departures for the supplied CRS code within a defined time window.
    # GetArrivalBoard(
    #    numRows: xsd:unsignedShort, crs: ns2:CRSType, filterCrs: ns2:CRSType, filterType: ns2:FilterType, timeOffset: xsd:int, timeWindow: xsd:int, _soapheaders = {
    #    AccessToken: ns0:AccessToken}) -> GetStationBoardResult: ns4:StationBoard
    def get_arrival_board(self, numRows, crs, filterCrs=None, filterType=None, timeOffset=0, timeWindow=120):

        self._validate_num_rows(numRows)
        self._validate_crs(crs)
        if filterCrs is not None:
            self._validate_filterCrs(filterCrs)
        if filterType is not None:
            self._validate_filterType(filterType)
        self._validate_timeOffset(timeOffset)
        self._validate_timeWindow(timeWindow)

        started = datetime.now()
        speed = None
        try:
            data = self._client.service.GetArrivalBoard(numRows=numRows,
                                                        crs=crs,
                                                        filterCrs=filterCrs,
                                                        filterType=filterType,
                                                        timeOffset=timeOffset,
                                                        timeWindow=timeWindow,
                                                        _soapheaders=self._header)
            speed = started - datetime.now()

            if data is not None:
                result = {"arrival_board": data}
                return self._create_success_payload("arrival_board", started, speed, result)

            return self._create_failure_payload("arrival_board", started, speed)

        except Exception as error:
            return self._create_exception_failure_payload("arrival_board", started, speed, error)

    # Returns all public arrivals and departures for the supplied CRS code within a defined time window.
    # GetArrivalDepartureBoard(
    #    numRows: xsd:unsignedShort, crs: ns2:CRSType, filterCrs: ns2:CRSType, filterType: ns2:FilterType, timeOffset: xsd:int, timeWindow: xsd:int, _soapheaders = {
    #    AccessToken: ns0:AccessToken}) -> GetStationBoardResult: ns4:StationBoard
    def get_arrival_and_departure_boards(self, numRows, crs, filterCrs=None, filterType=None, timeOffset=0, timeWindow=120):

        self._validate_num_rows(numRows)
        self._validate_crs(crs)
        if filterCrs is not None:
            self._validate_filterCrs(filterCrs)
        if filterType is not None:
            self._validate_filterType(filterType)
        self._validate_timeOffset(timeOffset)
        self._validate_timeWindow(timeWindow)

        started = datetime.now()
        speed = None
        try:
            data = self._client.service.GetArrivalDepartureBoard(numRows=numRows,
                                                                 crs=crs,
                                                                 filterCrs=filterCrs,
                                                                 filterType=filterType,
                                                                 timeOffset=timeOffset,
                                                                 timeWindow=timeWindow,
                                                                 _soapheaders=self._header)
            speed = started - datetime.now()

            if data is not None:
                result = {"arrival_and_departure_boards": data}
                return self._create_success_payload("arrival_and_departure_boards", started, speed, result)

            return self._create_failure_payload("arrival_and_departure_boards", started, speed)

        except Exception as error:
            return self._create_exception_failure_payload("arrival_and_departure_boards", started, speed, error)

    # Returns all public arrivals and departures for the supplied CRS code within a defined time window, including service details.
    # GetDepBoardWithDetails(
    #    numRows: xsd:unsignedShort, crs: ns2:CRSType, filterCrs: ns2:CRSType, filterType: ns2:FilterType, timeOffset: xsd:int, timeWindow: xsd:int, _soapheaders = {
    #    AccessToken: ns0:AccessToken}) -> GetStationBoardResult: ns4:StationBoardWithDetails
    def get_departure_board_with_details(self, numRows, crs, filterCrs=None, filterType=None, timeOffset=0, timeWindow=120):

        self._validate_num_rows(numRows)
        self._validate_crs(crs)
        if filterCrs is not None:
            self._validate_filterCrs(filterCrs)
        if filterType is not None:
            self._validate_filterType(filterType)
        self._validate_timeOffset(timeOffset)
        self._validate_timeWindow(timeWindow)

        started = datetime.now()
        speed = None
        try:
            data = self._client.service.GetDepBoardWithDetails(numRows=numRows,
                                                               crs=crs,
                                                               filterCrs=filterCrs,
                                                               filterType=filterType,
                                                               timeOffset=timeOffset,
                                                               timeWindow=timeWindow,
                                                               _soapheaders=self._header)
            speed = started - datetime.now()

            if data is not None:
                result = {"departure_board_with_details": data}
                return self._create_success_payload("departure_board_with_details", started, speed, result)

            return self._create_failure_payload("departure_board_with_details", started, speed)

        except Exception as error:
            return self._create_exception_failure_payload("departure_board_with_details", started, speed, error)

    # Returns all public departures for the supplied CRS code within a defined time window.
    # GetDepartureBoard(
    #    numRows: xsd:unsignedShort, crs: ns2:CRSType, filterCrs: ns2:CRSType, filterType: ns2:FilterType, timeOffset: xsd:int, timeWindow: xsd:int, _soapheaders = {
    #    AccessToken: ns0:AccessToken}) -> GetStationBoardResult: ns4:StationBoard
    def get_departure_board(self, numRows, crs, filterCrs=None, filterType=None, timeOffset=0, timeWindow=120):

        self._validate_num_rows(numRows)
        self._validate_crs(crs)
        if filterCrs is not None:
            self._validate_filterCrs(filterCrs)
        if filterType is not None:
            self._validate_filterType(filterType)
        self._validate_timeOffset(timeOffset)
        self._validate_timeWindow(timeWindow)

        started = datetime.now()
        speed = None
        try:
            data = self._client.service.GetDepartureBoard(numRows=numRows,
                                                          crs=crs,
                                                          filterCrs=filterCrs,
                                                          filterType=filterType,
                                                          timeOffset=timeOffset,
                                                          timeWindow=timeWindow,
                                                          _soapheaders=self._header)
            speed = started - datetime.now()

            if data is not None:
                result = {"departure_board": data}
                return self._create_success_payload("departure_board", started, speed, result)

            return self._create_failure_payload("departure_board", started, speed)

        except Exception as error:
            return self._create_exception_failure_payload("departure_board", started, speed, error)

    # Returns the public departure for the supplied CRS code within a defined time window to the locations specified in the filter with the earliest arrival time at the filtered location.
    # GetFastestDepartures(crs: ns2:CRSType, filterList: {crs: ns2:
    #    CRSType[]}, timeOffset: xsd:int, timeWindow: xsd:int, _soapheaders = {
    #    AccessToken: ns0:AccessToken}) -> DeparturesBoard: ns4:DeparturesBoard
    def get_fastest_departures(self, crs, filterList, timeOffset=0, timeWindow=120):

        self._validate_crs(crs)
        self._validate_filterList(filterList)
        self._validate_timeOffset(timeOffset)
        self._validate_timeWindow(timeWindow)

        started = datetime.now()
        speed = None
        try:
            data = self._client.service.GetFastestDepartures(crs=crs,
                                                             filterList=filterList,
                                                             timeOffset=timeOffset,
                                                             timeWindow=timeWindow,
                                                             _soapheaders=self._header)

            speed = started - datetime.now()

            if data is not None:
                result = {"fastest_departures": data}
                return self._create_success_payload("fastest_departures", started, speed, result)

            return self._create_failure_payload("fastest_departures", started, speed)

        except Exception as error:
            return self._create_exception_failure_payload("fastest_departures", started, speed, error)


    # Returns the public departure for the supplied CRS code within a defined time window to the locations specified in the filter with the earliest arrival time at the filtered location, including service details.
    # GetFastestDeparturesWithDetails(crs: ns2:CRSType, filterList: {crs: ns2:
    #    CRSType[]}, timeOffset: xsd:int, timeWindow: xsd:int, _soapheaders = {
    #    AccessToken: ns0:AccessToken}) -> DeparturesBoard: ns4:DeparturesBoardWithDetails
    def get_fastest_departures_with_details(self, crs, filterList, timeOffset=0, timeWindow=120):

        self._validate_crs(crs)
        self._validate_filterList(filterList)
        self._validate_timeOffset(timeOffset)
        self._validate_timeWindow(timeWindow)

        started = datetime.now()
        speed = None
        try:
            data = self._client.service.GetFastestDeparturesWithDetails(crs=crs,
                                                                        filterList=filterList,
                                                                        timeOffset=timeOffset,
                                                                        timeWindow=timeWindow,
                                                                        _soapheaders=self._header)
            speed = started - datetime.now()

            if data is not None:
                result = {"fastest_departures_with_details": data}
                return self._create_success_payload("fastest_departures_with_details", started, speed, result)

            return self._create_failure_payload("fastest_departures_with_details", started, speed)

        except Exception as error:
            return self._create_exception_failure_payload("fastest_departures_with_details", started, speed, error)

    # Returns the next public departure for the supplied CRS code within a defined time window to the locations specified in the filter.
    # GetNextDepartures(crs: ns2:CRSType, filterList: {crs: ns2:
    #    CRSType[]}, timeOffset: xsd:int, timeWindow: xsd:int, _soapheaders = {
    #    AccessToken: ns0:AccessToken}) -> DeparturesBoard: ns4:DeparturesBoard
    def get_next_departures(self, crs, filterList, timeOffset=0, timeWindow=120):

        self._validate_crs(crs)
        self._validate_filterList(filterList)
        self._validate_timeOffset(timeOffset)
        self._validate_timeWindow(timeWindow)

        started = datetime.now()
        speed = None
        try:
            data = self._client.service.GetNextDepartures(crs=crs,
                                                          filterList=filterList,
                                                          timeOffset=timeOffset,
                                                          timeWindow=timeWindow,
                                                          _soapheaders=self._header)

            speed = started - datetime.now()

            if data is not None:
                result = {"next_departures": data}
                return self._create_success_payload("next_departures", started, speed, result)

            return self._create_failure_payload("next_departures", started, speed)

        except Exception as error:
            return self._create_exception_failure_payload("next_departures", started, speed, error)

    # Returns the next public departure for the supplied CRS code within a defined time window to the locations specified in the filter, including service details.
    # GetNextDeparturesWithDetails(crs: ns2:CRSType, filterList: {crs: ns2:
    #    CRSType[]}, timeOffset: xsd:int, timeWindow: xsd:int, _soapheaders = {
    #    AccessToken: ns0:AccessToken}) -> DeparturesBoard: ns4:DeparturesBoardWithDetails
    def get_next_departures_with_details(self, crs, filterList, timeOffset=0, timeWindow=120):

        self._validate_crs(crs)
        self._validate_filterList(filterList)
        self._validate_timeOffset(timeOffset)
        self._validate_timeWindow(timeWindow)

        started = datetime.now()
        speed = None
        try:
            data = self._client.service.GetNextDeparturesWithDetails(crs=crs,
                                                                     filterList=filterList,
                                                                     timeOffset=timeOffset,
                                                                     timeWindow=timeWindow,
                                                                     _soapheaders=self._header)

            speed = started - datetime.now()

            if data is not None:
                result = {"next_departures_with_details": data}
                return self._create_success_payload("next_departures_with_details", started, speed, result)

            return self._create_failure_payload("next_departures_with_details", started, speed)

        except Exception as error:
            return self._create_exception_failure_payload("next_departures_with_details", started, speed, error)

    # Returns service details for a specific service identified by a station board. These details are supplied relative
    # to the station board from which the serviceID field value was generated. Service details are only available while
    # the service appears on the station board from which it was obtained. This is normally for two minutes after it is
    # expected to have departed, or after a terminal arrival. If a request is made for a service that is no longer
    # available then a null value is returned.
    # GetServiceDetails(serviceID: ns3:ServiceIDType, _soapheaders = {
    #    AccessToken: ns0:AccessToken}) -> GetServiceDetailsResult: ns4:ServiceDetails
    def get_service_details(self, serviceID):

        self._validate_serviceID(serviceID)

        started = datetime.now()
        speed = None
        try:
            data = self._client.service.GetServiceDetails(serviceID=serviceID,
                                                          _soapheaders=self._header)

            speed = started - datetime.now()

            if data is not None:
                result = {"service_details": data}
                return self._create_success_payload("next_departures_with_details", started, speed, result)

            return self._create_failure_payload("service_details", started, speed)

        except Exception as error:
            return self._create_exception_failure_payload("service_details", started, speed, error)

    def get_station_name_from_code(self, code):
        started = datetime.now()
        speed = None

        if code in self._stations_to_codes:
            result = {"station_name": self._stations_to_codes[code]}
            speed = started - datetime.now()
            return self._create_success_payload("station_name_from_code", started, speed, result)

        return self._create_failure_payload("station_name_from_code", started, speed)

    def get_station_code_from_name(self, name):
        started = datetime.now()
        speed = None

        name = name.upper()
        if name in self._codes_to_stations:
            result = {"station_code": self._codes_to_stations[name]}
            return self._create_success_payload("get_station_code_from_name", started, speed, result)

        return self._create_failure_payload("get_station_code_from_name", started, speed)

    def get_station_code_from_fuzzy_name(self, name, limit=3, threshold=80):
        if name in self._codes_to_stations:
            return self._codes_to_stations[name]

        codes = []
        matches = self._match_station(name, limit, threshold)
        for name in matches:
            codes.append(self._codes_to_stations[name])

        if codes:
            return ", ".join(codes)

        return None

    def _match_station(self, candidate, limit=3, threshold=80):
        results = process.extract(candidate, self._stations_cache, limit=limit)
        matches = []
        for station, match in results:
            if match >= threshold:
                matches.append(station)
        return matches

    def _response_to_json(self, api, response):
        return response

    def next_trains_from_station(self, station, platform=None, origin=None, destination=None):

        started = datetime.now()
        speed = None

        try:
            if len(station) != 3:
                response = self.get_station_code_from_name(station)
                station = response['response']['payload']['station_code']

            if station is None:
                raise Exception ("Unknown station")

            if origin:
                if len(origin) != 3:
                    response = self.get_station_code_from_name(origin)
                    origin = response['response']['payload']['station_code']

            if destination:
                if len(destination) != 3:
                    response = self.get_station_code_from_name(destination)
                    destination = response['response']['payload']['station_code']

            departure_board = self.get_departure_board(10, station)

            response = departure_board.get('response')
            if response is None:
                raise Exception("no response in get_departure_board")

            if response.get('status', 'failure') == 'failure':
                raise Exception("Service called failed")

            payload = response.get('payload')
            if payload is None:
                raise Exception("no payload in get_departure_board")

            data = payload.get('departure_board')
            if data is None:
                raise Exception("no departure_board in get_departure_board")

            trainServices = data['trainServices']
            if trainServices is None:
                raise Exception("no trainServices in get_departure_board")

            services = trainServices.service
            if not services:
                raise Exception("no services in get_departure_board")

            result = {'platforms': {}}

            for service in services:
                service_operator = service.operator
                service_std = service.std
                service_etd = service.etd
                service_platform = service.platform
                service_origin = service.origin.location[0].locationName
                service_origin_crs = service.origin.location[0].crs
                service_destination = service.destination.location[0].locationName
                service_destination_crs = service.destination.location[0].crs

                if platform is not None:
                    if platform != service_platform:
                        continue

                if origin is not None:
                    if origin != service_origin_crs:
                        continue

                if destination is not None:
                    if destination != service_destination_crs:
                        continue

                if service_platform not in result.get('platforms'):
                    result['platforms'][service_platform] = []

                result['platforms'][service_platform].append({'operator': service_operator,
                                                              'std': service_std,
                                                              'eta': service_etd,
                                                              'origin': service_origin,
                                                              'destination': service_destination})

            speed = started - datetime.now()

            return self._create_success_payload("next_trains_from_station", started, speed, {'departures': result})

        except Exception as error:
            return self._create_exception_failure_payload("next_trains_from_station", started, speed, error)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from abc import ABC
from zeep import Client
from programy.services.base import Service
from programy.services.base import ServiceException


class WSDLServiceException(ServiceException):

    def __init__(self, msg):
        ServiceException.__init__(self, msg)


class WSDLService(Service, ABC):

    def __init__(self, configuration):
        Service.__init__(self, configuration)

    def create_wsdl_client(self, wsdl):
        self._client = Client(wsdl)

    def _response_to_json(self, api, response):
        raise NotImplementedError()     # pragma: no cover

    def _add_base_payload(self, data, status, started, speed):
        if started is not None:
            data['response']['started'] = started.strftime("%d/%m/%Y, %H:%M:%S")
        if speed is not None:
            data['response']['speed'] = str(speed.microseconds/1000) + "ms"
        data['response']['status'] = status
        data['response']['service'] = self.name
        data['response']['category'] = self.category

    def _create_success_payload(self, api, started, speed, response):
        data = {}
        data['response'] = {}
        self._add_base_payload(data, "success", started, speed)
        data['response']['payload'] = self._response_to_json(api, response)
        data['response']['api'] = api
        return data

    def _create_failure_payload(self, api, started, speed):
        data = {}
        data['api'] = api
        data['response'] = {}
        data['response']['api'] = api
        self._add_base_payload(data, "failure", started, speed)
        data['response']['payload'] = {}
        data['response']['payload']['type'] = 'general'
        return data

    def _create_exception_failure_payload(self, api, started, speed, err):
        data = {}
        data['response'] = {}
        data['response']['api'] = api
        self._add_base_payload(data, "failure", started, speed)
        data['response']['payload'] = {}
        data['response']['payload']['type'] = 'general'
        data['response']['payload']['error'] = str(err)
        return data




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.config.file.xml_file import XMLConfigurationFile
from programy.config.file.yaml_file import YamlConfigurationFile
from programy.config.file.json_file import JSONConfigurationFile
from programy.utils.substitutions.substitues import Substitutions


class ConfigurationFactory:

    @classmethod
    def load_configuration_from_file(cls, client_configuration, filename, file_format=None, bot_root=".",
                                     subs: Substitutions = None):

        if file_format is None or not file_format:
            file_format = ConfigurationFactory.guess_format_from_filename(filename)

        config_file = ConfigurationFactory.get_config_by_name(file_format)
        return config_file.load_from_file(filename, client_configuration, bot_root, subs)

    @classmethod
    def guess_format_from_filename(cls, filename):
        if "." not in filename:
            raise Exception("No file extension to allow format guessing!")

        last_dot = filename.rfind(".")
        file_format = filename[last_dot + 1:]
        return file_format

    @classmethod
    def get_config_by_name(cls, file_format):
        file_format = file_format.lower()

        if file_format == 'yaml':
            return YamlConfigurationFile()
        elif file_format == 'json':
            return JSONConfigurationFile()
        elif file_format == 'xml':
            return XMLConfigurationFile()
        else:
            raise Exception("Unsupported configuration format:", file_format)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.clients.config import ClientConfigurationData
from programy.utils.substitutions.substitues import Substitutions


class TelegramConfiguration(ClientConfigurationData):

    def __init__(self):
        ClientConfigurationData.__init__(self, "telegram")
        self._unknown_command = None
        self._unknown_command_srai = None

    @property
    def unknown_command(self):
        return self._unknown_command

    @property
    def unknown_command_srai(self):
        return self._unknown_command_srai

    def load_configuration_section(self, configuration_file, section, bot_root, subs: Substitutions = None):
        assert section is not None

        self._unknown_command = configuration_file.get_option(section, "unknown_command",
                                                              missing_value="Unknown command", subs=subs)
        self._unknown_command_srai = configuration_file.get_option(section, "unknown_command_srai",
                                                                   missing_value=None, subs=subs)
        super(TelegramConfiguration, self).load_configuration_section(configuration_file, section, bot_root,
                                                                      subs=subs)

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['unknown_command'] = "Sorry, that is not a command I have been taught yet!"
            data['unknown_command_srai'] = 'YTELEGRAM_UNKNOWN_COMMAND'
        else:
            data['unknown_command'] = self._unknown_command
            data['unknown_command_srai'] = self._unknown_command_srai

        super(TelegramConfiguration, self).to_yaml(data, defaults)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import urllib.parse
import os
from programy.utils.logging.ylogger import YLogger
from programy.services.base import ServiceQuery
from programy.services.rest.base import RESTService
from programy.services.rest.base import RESTServiceException


class AccuWeatherPostCodeSearchServiceQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return AccuWeatherTextSearchServiceQuery(service)

    def parse_matched(self, matched):
        self._postcode = ServiceQuery._get_matched_var(matched, 0, "postcode")
        self._country = ServiceQuery._get_matched_var(matched, 1, "country", optional=True)

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._postcode = None
        self._country = None

    def execute(self):
        return self._service.postcodesearch(self._postcode, self._country)

    def aiml_response(self, response):
        result = "KEY {0}".format(AccuWeatherService.get_location_key(response))
        YLogger.debug(self, result)
        return result


class AccuWeatherTextSearchServiceQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return AccuWeatherTextSearchServiceQuery(service)

    def parse_matched(self, matched):
        self._location = ServiceQuery._get_matched_var(matched, 0, "location")
        self._country = ServiceQuery._get_matched_var(matched, 1, "country", optional=True)

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._location = None
        self._country = None

    def execute(self):
        return self._service.textsearch(self._location, self._country)

    def aiml_response(self, response):
        result = "KEY {0}".format(AccuWeatherService.get_location_key(response))
        YLogger.debug(self, result)
        return result


class AccuWeatherConditionsServiceQuery(ServiceQuery):

    @staticmethod
    def create(service):
        return AccuWeatherConditionsServiceQuery(service)

    def parse_matched(self, matched):
        self._location = ServiceQuery._get_matched_var(matched, 0, "location")

    def __init__(self, service):
        ServiceQuery.__init__(self, service)
        self._location = None

    def execute(self):
        return self._service.conditions(self._location)

    def aiml_response(self, response):
        try:
            payload = response['response']['payload'][0]
            result = "EPOC {0} WEATHERTEXT {1} HASPRECIPITATION {2} PRECIPITATIONTYPE {3} ISDAYTIME {4} TEMP {5} UNIT {6}".format(
                    payload['EpochTime'],
                    payload['WeatherText'],
                    payload['HasPrecipitation'],
                    payload['PrecipitationType'],
                    payload['IsDayTime'],
                    payload['Temperature']['Metric']['Value'],
                    payload['Temperature']['Metric']['Unit'])
            YLogger.debug(self, result)
            return result

        except Exception as error:
            YLogger.exception(self, "Failed to parse Accuweather response to AIML", error)

        return None


class AccuWeatherServiceException(RESTServiceException):

    def __init__(self, msg):
        RESTServiceException.__init__(self, msg)


class AccuWeatherService(RESTService):
    """
    """
    PATTERNS = [
        [r"TEXTSEARCH\sLOCATION\s(.+)", AccuWeatherTextSearchServiceQuery],
        [r"CONDITIONS\sLOCATION\s(.+)", AccuWeatherConditionsServiceQuery]
    ]

    POSTCODESEARCH_BASE_URL="http://dataservice.accuweather.com/locations/v1/postalcodes/search"
    POSTCODESEARCH_COUNTRY_BASE_URL="http://dataservice.accuweather.com/locations/v1/postalcodes/{0}/search"
    TEXTSEARCH_BASE_URL="http://dataservice.accuweather.com/locations/v1/search"
    TEXTSEARCH_COUNTRY_BASE_URL="http://dataservice.accuweather.com/locations/v1/{0}/search"
    CONDITIONS_BASE_URL="http://dataservice.accuweather.com/currentconditions/v1/"

    def __init__(self, configuration):
        RESTService.__init__(self, configuration)
        self._api_key = None

    def patterns(self) -> list:
        return AccuWeatherService.PATTERNS

    def initialise(self, client):
        self._api_key = client.license_keys.get_key('ACCUWEATHER_APIKEY')
        if self._api_key is None:
            YLogger.error(self, "ACCUWEATHER_APIKEY missing from license.keys, service will not function correctly!")

    def get_default_aiml_file(self):
        return os.path.dirname(__file__) + os.sep + "accuweather.aiml"

    @staticmethod
    def get_default_conf_file():
        return os.path.dirname(__file__) + os.sep + "accuweather.conf"

    def _build_postcodesearch_url(self, postcode, country, language=None, details=None, offset=None, alias=None):
        if country is None:
            url = AccuWeatherService.POSTCODESEARCH_BASE_URL
        else:
            url = AccuWeatherService.POSTCODESEARCH_COUNTRY_BASE_URL.format(country)
        url += "?apikey={0}".format(self._api_key)
        url += "&q={0}".format(postcode)
        if language is not None:
            url += "&language={0}".format(language)
        if details is not None:
            url += "&details={0}".format(details)
        if offset is not None:
            url += "&offset={0}".format(offset)
        if alias is not None:
            url += "&alias={0}".format(alias)
        return url

    def postcodesearch(self, postcode, country=None, language=None, details=None, offset=None, alias=None):
        url = self._build_postcodesearch_url(urllib.parse.quote_plus(postcode), country, language, details, offset, alias)
        response = self.query('postcodesearch', url)
        return response

    def _build_textsearch_url(self, postcode, country, language=None, details=None, offset=None, alias=None):
        if country is None:
            url = AccuWeatherService.TEXTSEARCH_BASE_URL
        else:
            url = AccuWeatherService.TEXTSEARCH_COUNTRY_BASE_URL.format(country)
        url += "?apikey={0}".format(self._api_key)
        url += "&q={0}".format(postcode)
        if language is not None:
            url += "&language={0}".format(language)
        if details is not None:
            url += "&details={0}".format(details)
        if offset is not None:
            url += "&offset={0}".format(offset)
        if alias is not None:
            url += "&alias={0}".format(alias)
        return url

    def textsearch(self, location, country=None, language=None, details=None, offset=None, alias=None):
        url = self._build_textsearch_url(urllib.parse.quote_plus(location), country, language, details, offset, alias)
        response = self.query('textsearch', url)
        return response

    def _build_conditions_url(self, location, language=None, details=None):
        url = AccuWeatherService.CONDITIONS_BASE_URL
        url += location
        url += "?apikey={0}".format(self._api_key)
        if language is not None:
            url += "&language={0}".format(language)
        if details is not None:
            url += "&details={0}".format(details)
        return url

    def conditions(self, location, language=None, details=None):
        url = self._build_conditions_url(location, language, details)
        response = self.query('conditions', url)
        return response

    def _response_to_json(self, api, response):
        return response.json()

    @staticmethod
    def get_location_key(data):
        response = data.get("response")
        if response is not None:
            payload = response.get("payload")
            if payload is not None:
                if len(payload) > 0:
                    payload0 = payload[0]
                    key = payload0.get("Key")
                    return key
        return None


"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import redis
from programy.utils.logging.ylogger import YLogger
from programy.storage.engine import StorageEngine
from programy.storage.stores.nosql.redis.store.conversations import RedisConversationStore


class RedisStorageEngine(StorageEngine):

    def __init__(self, configuration):
        StorageEngine.__init__(self, configuration)
        self._prefix = "programy"  # Default value
        self._sessions_set_key = self._create_sessions_set_key()
        self._redis = None

    @property
    def redis(self):
        return self._redis

    @property
    def prefix(self):
        return self._prefix

    @property
    def sessions_set_key(self):
        return self._sessions_set_key

    def _create_sessions_set_key(self):
        return "{prefix}:sessions".format(prefix=self._prefix)

    def initialise(self):
        self._prefix = self.configuration.prefix
        self._sessions_set_key = self._create_sessions_set_key()

        if self.configuration.password is not None:
            self._redis = redis.StrictRedis(
                host=self.configuration.host,
                port=self.configuration.port,
                password=self.configuration.password,
                db=self.configuration.db)
        else:
            self._redis = redis.StrictRedis(
                host=self.configuration.host,
                port=self.configuration.port,
                db=self.configuration.db)

        if self.configuration.drop_all_first is True:
            try:
                self.conversation_store().empty()
            except Exception as excep:
                YLogger.exception(self, "Failed deleting conversation redis data - ", excep)

    def conversation_store(self):
        return RedisConversationStore(self)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.parser.exceptions import ParserException
from programy.parser.pattern.nodes.base import PatternNode
from programy.parser.pattern.equalsmatch import EqualsMatch


class PatternWordNode(PatternNode):

    def __init__(self, word, userid='*'):
        PatternNode.__init__(self, userid)
        self._word = word

    def is_word(self):
        return True

    @property
    def word(self):
        return self._word

    def to_xml(self, client_context, include_user=False):
        string = ""
        if include_user:
            string += '<word userid="%s" word="%s">' % (self.userid, self.word)
        else:
            string += '<word word="%s">' % self.word
        string += super(PatternWordNode, self).to_xml(client_context, include_user)
        string += '</word>\n'
        return string

    def to_string(self, verbose=True):
        if verbose is True:
            return "WORD [%s] [%s] word=[%s]" % (self.userid, self._child_count(verbose), self.word)
        return "WORD [%s]" % (self.word)

    def can_add(self, new_node):
        if new_node.is_root():
            raise ParserException("Cannot add root node to child node")

    def equivalent(self, other):
        if other.is_word():
            if self.userid == other.userid:
                if self._word == other.word:
                    return True
        return False

    def equals(self, client_context, words, word_no):
        word = words.word(word_no)

        if self.userid != '*':
            if self.userid != client_context.userid:
                return EqualsMatch(False, word_no)

        return EqualsMatch(self.equals_ignore_case(self._word, word), word_no, word)
